# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

"""
Contrast adjustment
^^^^^^^^^^^^^^^^^^^

The `contrast adjustment` panel is a widget which displays the image levels
histogram and allows to manipulate it in order to adjust the image contrast.

.. autoclass:: ContrastAdjustment
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from guidata.configtools import get_icon, get_image_layout
from guidata.dataset import DataSet, FloatItem
from guidata.qthelpers import add_actions, create_action, is_dark_theme
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from plotpy.config import CONF, _
from plotpy.constants import ID_CONTRAST, Y_LEFT, Y_RIGHT, PlotType
from plotpy.interfaces import IPanel, IVoiImageItemType
from plotpy.items import HistogramItem, XRangeSelection
from plotpy.lutrange import lut_range_threshold
from plotpy.panels.base import PanelWidget
from plotpy.plot.base import BasePlot, BasePlotOptions
from plotpy.styles import CurveParam, HistogramParam
from plotpy.tools import AntiAliasingTool, BasePlotMenuTool, SelectPointTool, SelectTool

if TYPE_CHECKING:
    from collections.abc import Callable, Generator

    from qtpy.QtGui import QCloseEvent
    from qtpy.QtWidgets import QWidget

    from plotpy.items import BaseImageItem
    from plotpy.plot.manager import PlotManager


class LevelsHistogram(BasePlot):
    """Image levels histogram widget

    Args:
        parent: parent widget
    """

    DEFAULT_NBINS = 256  # Default number of bins for histogram

    #: Signal emitted by LevelsHistogram when LUT range of some items was changed.
    #: For now, this signal is private. The public counterpart is emitted by
    #: the base plot class (see :py:attr:`.BasePlot.SIG_LUT_CHANGED`).
    SIG_VOI_CHANGED = QC.Signal()

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(
            parent=parent,
            options=BasePlotOptions(title="", section="histogram", type="curve"),
        )
        self.antialiased = False

        # a dict of dict : plot -> selected items -> HistogramItem
        self._tracked_items: dict[BasePlot, dict[BaseImageItem, HistogramItem]] = {}
        self.param = CurveParam(_("Curve"), icon="curve.png")
        self.param.read_config(CONF, "histogram", "curve")

        self.range = XRangeSelection(0, 1)
        self.range_mono_color = self.range.shapeparam.sel_line.color
        self.range_multi_color = CONF.get("histogram", "range/multi/color", "red")

        self.add_item(self.range, z=5)
        self.SIG_RANGE_CHANGED.connect(self.range_changed)
        self.set_active_item(self.range)

        self.setMinimumHeight(80)
        self.setAxisMaxMajor(Y_LEFT, 5)
        self.setAxisMaxMinor(Y_LEFT, 0)

        if parent is None:
            self.set_axis_title("bottom", "Levels")

    def connect_plot(self, plot: BasePlot) -> None:
        """Connect plot to histogram widget

        Args:
            plot: plot to connect to
        """
        if plot.options.type == PlotType.CURVE:
            # Connecting only to image plot widgets (allow mixing image and
            # curve widgets for the same plot manager -- e.g. in pyplot)
            return
        self.SIG_VOI_CHANGED.connect(plot.notify_colormap_changed)
        plot.SIG_ITEM_SELECTION_CHANGED.connect(self.selection_changed)
        plot.SIG_ITEM_REMOVED.connect(self.item_removed)
        plot.SIG_ACTIVE_ITEM_CHANGED.connect(self.active_item_changed)
        plot.SIG_AXIS_PARAMETERS_CHANGED.connect(
            lambda axis_id: self.axis_parameters_changed(plot=plot, axis_id=axis_id)
        )

    def tracked_items_gen(
        self,
    ) -> Generator[tuple[BaseImageItem, HistogramItem], None, None]:
        """Generator of tracked items"""
        for _plot, items in list(self._tracked_items.items()):
            for item_curve_tuple in list(items.items()):
                yield item_curve_tuple  # tuple item,curve

    def __del_known_items(self, known_items: dict, items: list) -> None:
        """Delete known items

        Args:
            known_items: dict of known items
            items: list of items to delete
        """
        del_curves = []
        for item in list(known_items.keys()):
            if item not in items:
                curve = known_items.pop(item)
                del_curves.append(curve)
        self.del_items(del_curves)

    def selection_changed(self, plot: BasePlot) -> None:
        """Selection changed callback

        Args:
            plot: plot whose selection changed
        """
        items: list[BaseImageItem] = plot.get_selected_items(
            item_type=IVoiImageItemType
        )
        known_items = self._tracked_items.setdefault(plot, {})

        if items:
            self.__del_known_items(known_items, items)
            if len(items) == 1:
                # Removing any cached item for other plots
                for other_plot, _items in list(self._tracked_items.items()):
                    if other_plot is not plot:
                        if not other_plot.get_selected_items(
                            item_type=IVoiImageItemType
                        ):
                            other_known_items = self._tracked_items[other_plot]
                            self.__del_known_items(other_known_items, [])
        else:
            # if all items are deselected we keep the last known
            # selection (for one plot only)
            for other_plot, _items in list(self._tracked_items.items()):
                if other_plot.get_selected_items(item_type=IVoiImageItemType):
                    self.__del_known_items(known_items, [])
                    break

        for item in items:
            if item not in known_items:
                histparam = HistogramParam(_("Histogram"), icon="histogram.png")
                histparam.logscale = False
                histparam.n_bins = self.DEFAULT_NBINS
                curve = HistogramItem(self.param, histparam, keep_weakref=True)
                curve.set_hist_source(item)
                self.add_item(curve, z=0)
                known_items[item] = curve

        nb_selected = len(list(self.tracked_items_gen()))
        if not nb_selected:
            self.replot()
            return
        self.param.shade = 1.0 / nb_selected
        for item, curve in self.tracked_items_gen():
            if np.issubdtype(item.data.dtype, np.integer):
                # For integer data, we use the full range of data type
                info = np.iinfo(item.data.dtype)
                curve.histparam.bin_min = info.min
                curve.histparam.bin_max = info.max
                curve.histparam.n_bins = min(
                    info.max - info.min + 1, self.DEFAULT_NBINS
                )
            else:
                curve.histparam.bin_min = None
                curve.histparam.bin_max = None
                curve.histparam.n_bins = self.DEFAULT_NBINS
            self.param.update_item(curve)
            curve.histparam.update_hist(curve)

        self.active_item_changed(plot)

        # Rescaling histogram plot axes for better visibility
        ymax = None
        for item in known_items:
            curve = known_items[item]
            _x, y = curve.get_data()
            ymax0 = y.mean() + 3 * y.std()
            if ymax is None or ymax0 > ymax:
                ymax = ymax0
        ymin, _ymax = self.get_axis_limits("left")
        if ymax is not None:
            self.set_axis_limits("left", ymin, ymax)
            self.replot()

    def item_removed(self, item: BaseImageItem) -> None:
        """Item removed callback

        Args:
            item: item which was removed
        """
        replot = False
        for _plot, items_dict in list(self._tracked_items.items()):
            if item in items_dict:
                try:
                    self.del_item(items_dict[item])
                    replot = True
                except ValueError:
                    pass  # Histogram has not yet been created
                items_dict.pop(item)
                break
        if replot:
            self.replot()

    def active_item_changed(self, plot: BasePlot) -> None:
        """Active item changed callback

        Args:
            plot: plot whose active item changed
        """
        items = plot.get_selected_items(item_type=IVoiImageItemType)
        if not items:
            return

        active = plot.get_last_active_item(IVoiImageItemType)
        if active:
            active_range = active.get_lut_range()
        else:
            active_range = None

        multiple_ranges = False
        for item, curve in self.tracked_items_gen():
            if active_range != item.get_lut_range():
                multiple_ranges = True
        if active_range is not None:
            _m, _M = active_range
            self.set_range_style(multiple_ranges)
            self.range.set_range(_m, _M, dosignal=False)
        self.replot()

    def axis_parameters_changed(self, plot: BasePlot, axis_id: int) -> None:
        """Plot axis parameters changed callback

        Args:
            axis_id: axis ID
        """
        if axis_id == Y_RIGHT:
            # Colormap bounds changed, we need to update the range accordingly:
            self.active_item_changed(plot)

    def set_range_style(self, multiple_ranges: bool) -> None:
        """Set range style

        Args:
            multiple_ranges: whether multiple ranges are selected
        """
        if multiple_ranges:
            self.range.shapeparam.sel_line.color = self.range_multi_color
        else:
            self.range.shapeparam.sel_line.color = self.range_mono_color
        self.range.shapeparam.update_item(self.range)

    def set_range(self, zmin: float, zmax: float) -> bool:
        """Set range

        Args:
            zmin: minimum value
            zmax: maximum value

        Returns:
            True if range was changed, False otherwise
        """
        if zmin < zmax:
            self.set_range_style(False)
            self.range.set_range(zmin, zmax)
            self.replot()
            return True
        else:
            # Range was not changed
            return False

    def range_changed(
        self, _rangesel: XRangeSelection, zmin: float, zmax: float
    ) -> None:
        """Range changed callback

        Args:
            _rangesel: range selection
            zmin: minimum value
            zmax: maximum value
        """
        for item, _curve in self.tracked_items_gen():
            item.set_lut_range([zmin, zmax])
        self.SIG_VOI_CHANGED.emit()

    @staticmethod
    def apply_full_range_func(
        item: BaseImageItem, _curve: HistogramItem
    ) -> tuple[float, float]:
        """Apply full range function

        Args:
            item: item to apply full range to
            _curve: histogram curve associated with item

        Returns:
            tuple of minimum and maximum values
        """
        return item.get_lut_range_full()

    @staticmethod
    def apply_min_func(
        item: BaseImageItem, _curve: HistogramItem, zmin: float
    ) -> tuple[float, float]:
        """Apply minimum function

        Args:
            item: item to apply minimum to
            _curve: histogram curve associated with item
            zmin: minimum value

        Returns:
            tuple of minimum and maximum values
        """
        _zmin, zmax = item.get_lut_range()
        return zmin, zmax

    @staticmethod
    def apply_max_func(
        item: BaseImageItem, _curve: HistogramItem, zmax: float
    ) -> tuple[float, float]:
        """Apply maximum function

        Args:
            item: item to apply maximum to
            _curve: histogram curve associated with item
            zmax: maximum value

        Returns:
            tuple of minimum and maximum values
        """
        zmin, _zmax = item.get_lut_range()
        return zmin, zmax

    @staticmethod
    def reduce_range_func(
        item: BaseImageItem, curve: HistogramItem, percent: float
    ) -> tuple[float, float]:
        """Reduce range function

        Args:
            item: item to reduce range of
            curve: histogram curve associated with item
            percent: percentage of range to reduce

        Returns:
            tuple of minimum and maximum values
        """
        return lut_range_threshold(item, curve.bins, percent)

    def apply_range_function(self, func: Callable, *args, **kwargs) -> None:
        """Apply range function

        Args:
            func: function to apply
            *args: arguments to pass to function
            **kwargs: keyword arguments to pass to function
        """
        item = None
        zmin, zmax = np.inf, -np.inf
        for item, curve in self.tracked_items_gen():
            zmin_i, zmax_i = func(item, curve, *args, **kwargs)
            zmin = min(zmin, zmin_i)
            zmax = max(zmax, zmax_i)
        if zmin != np.inf and zmax != -np.inf:
            self.set_range(zmin, zmax)

    def set_full_range(self) -> None:
        """Set range bounds to image min/max levels"""
        self.apply_range_function(self.apply_full_range_func)

    def eliminate_outliers(self, percent: float) -> None:
        """Eliminate outliers

        Args:
            percent: percentage of outliers to eliminate (eliminate percent/2
             on each side)
        """
        self.apply_range_function(self.reduce_range_func, percent)

    def set_min(self, zmin: float) -> None:
        """Set minimum value

        Args:
            zmin: minimum value
        """
        self.apply_range_function(self.apply_min_func, zmin)

    def set_max(self, zmax: float) -> None:
        """Set maximum value

        Args:
            zmax: maximum value
        """
        self.apply_range_function(self.apply_max_func, zmax)


class EliminateOutliersParam(DataSet):
    percent = FloatItem(
        _("Eliminate outliers") + " (%)", default=2.0, min=0.0, max=100.0 - 1e-6
    )


class ContrastAdjustment(PanelWidget):
    """Contrast adjustment tool

    Args:
        parent: parent widget
    """

    __implements__ = (IPanel,)
    PANEL_ID = ID_CONTRAST
    PANEL_TITLE = _("Contrast")
    PANEL_ICON = "contrast.png"

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.local_manager = None  # local manager for the histogram plot
        self.manager = None  # manager for the associated image plot

        # Storing min/max markers for each active image
        self.min_markers = {}
        self.max_markers = {}

        # Select point tools
        self.min_select_tool = None
        self.max_select_tool = None

        color = "#bbbbbb" if is_dark_theme() else "#444444"
        style = "<span style='color: %s'><b>{}</b></span>" % color
        layout, _label = get_image_layout(
            self.PANEL_ICON, style.format(self.PANEL_TITLE), alignment=QC.Qt.AlignLeft
        )
        layout.setAlignment(QC.Qt.AlignCenter)
        vlayout = QW.QVBoxLayout()
        vlayout.addLayout(layout)

        # Avoid circular import
        # pylint-disable=import-outside-toplevel
        from plotpy.plot.manager import PlotManager

        self.local_manager = PlotManager(self)
        self.histogram = LevelsHistogram(parent)
        vlayout.addWidget(self.histogram)
        self.local_manager.add_plot(self.histogram)
        hlayout = QW.QHBoxLayout()
        self.setLayout(hlayout)
        hlayout.addLayout(vlayout)

        self.toolbar = toolbar = QW.QToolBar(self)
        toolbar.setIconSize(QC.QSize(12, 12))
        toolbar.setOrientation(QC.Qt.Horizontal)
        toolbar.setToolButtonStyle(QC.Qt.ToolButtonTextBesideIcon)
        layout.addWidget(toolbar)

        # Add standard plot-related tools to the local manager
        lman = self.local_manager
        lman.add_tool(SelectTool)
        lman.add_tool(BasePlotMenuTool, "item")
        lman.add_tool(BasePlotMenuTool, "axes")
        lman.add_tool(BasePlotMenuTool, "grid")
        lman.add_tool(AntiAliasingTool)
        lman.get_default_tool().activate()

        self.outliers_param = EliminateOutliersParam(self.PANEL_TITLE)

    def register_panel(self, manager: PlotManager) -> None:
        """Register panel to plot manager

        Args:
            manager: plot manager to register to
        """
        self.manager = manager
        default_toolbar = self.manager.get_default_toolbar()
        self.manager.add_toolbar(self.toolbar, "contrast")
        self.manager.set_default_toolbar(default_toolbar)
        self.setup_actions()
        for plot in manager.get_plots():
            self.histogram.connect_plot(plot)

    def configure_panel(self) -> None:
        """Configure panel"""
        self.min_select_tool = self.manager.add_tool(
            SelectPointTool,
            title=_("Min."),
            on_active_item=True,
            mode="create",
            tip=_("Select minimum level on image"),
            toolbar_id="contrast",
            end_callback=self.apply_min_selection,
            switch_to_default_tool=True,
        )
        self.max_select_tool = self.manager.add_tool(
            SelectPointTool,
            title=_("Max."),
            on_active_item=True,
            mode="create",
            tip=_("Select maximum level on image"),
            toolbar_id="contrast",
            end_callback=self.apply_max_selection,
            switch_to_default_tool=True,
        )

    def get_plot(self) -> BasePlot:
        """Get active plot

        Returns:
            active plot
        """
        return self.manager.get_active_plot()

    def closeEvent(self, event: QCloseEvent) -> None:
        """Reimplement Qt method"""
        self.hide()
        event.ignore()

    def setup_actions(self) -> None:
        """Setup actions"""
        fullrange_ac = create_action(
            self,
            _("Full range"),
            icon=get_icon("full_range.png"),
            triggered=self.histogram.set_full_range,
            tip=_("Scale the image's display range according to data range"),
        )
        outliers_ac = create_action(
            self,
            _("Eliminate %"),
            icon=get_icon("eliminate_outliers.png"),
            triggered=self.eliminate_outliers,
            tip=_(
                "Eliminate outliers from the image's display range, "
                "based on a percentage of the data range"
            ),
        )
        setrange_ac = create_action(
            self,
            _("Set range"),
            icon=get_icon("xrange.png"),
            triggered=self.set_range_manually,
            tip=_("Set the image's display range manually"),
        )
        add_actions(self.toolbar, [fullrange_ac, outliers_ac, setrange_ac])

    def eliminate_outliers(self) -> None:
        """Eliminate outliers"""

        def apply_callback(param: EliminateOutliersParam) -> None:
            """Apply callback"""
            self.histogram.eliminate_outliers(param.percent)

        zmin_0, zmax_0 = self.histogram.range.get_range()
        if self.outliers_param.edit(self, apply=apply_callback):
            apply_callback(self.outliers_param)
        else:
            self.histogram.set_range(zmin_0, zmax_0)

    def set_range_manually(self) -> None:
        """Set range manually"""

        class RangeParam(DataSet):
            """Range parameters"""

            zmin = FloatItem(_("Min. level"))
            zmax = FloatItem(_("Max. level"))

        def apply_callback(param: RangeParam) -> None:
            """Apply callback"""
            self.histogram.set_range(param.zmin, param.zmax)

        param = RangeParam(_("Set range"))
        zmin_0, zmax_0 = param.zmin, param.zmax = self.histogram.range.get_range()
        if param.edit(self, apply=apply_callback):
            apply_callback(param)
        else:
            self.histogram.set_range(zmin_0, zmax_0)

    def apply_min_selection(self, tool: SelectPointTool) -> None:
        """Apply minimum selection

        Args:
            tool: select point tool
        """
        item: BaseImageItem = self.get_plot().get_last_active_item(IVoiImageItemType)
        point = self.min_select_tool.get_coordinates()
        z = item.get_data(*point)
        self.histogram.set_min(z)
        tool.SIG_TOOL_JOB_FINISHED.emit()

    def apply_max_selection(self, tool: SelectPointTool) -> None:
        """Apply maximum selection

        Args:
            tool: select point tool
        """
        item: BaseImageItem = self.get_plot().get_last_active_item(IVoiImageItemType)
        point = self.max_select_tool.get_coordinates()
        z = item.get_data(*point)
        self.histogram.set_max(z)
        tool.SIG_TOOL_JOB_FINISHED.emit()

    def set_range(self, zmin: float, zmax: float) -> None:
        """Set contrast panel's histogram range

        Args:
            zmin: minimum value
            zmax: maximum value
        """
        self.histogram.set_range(zmin, zmax)
        # Update the levels histogram in case active item data has changed:
        self.histogram.selection_changed(self.get_plot())


assert_interfaces_valid(ContrastAdjustment)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
PlotPy plot benchmark
---------------------

This script benchmarks PlotPy plotting features.


Results obtained with PlotPy v2.3.5 on Windows 11 with a i5-1335U CPU @ 1.30 GHz:

.. code-block:: none

    PlotPy plot benchmark [Python 3.12.3 64 bits, Qt 5.15.2, PyQt 5.15.10 on Windows]

            N | ∆t (ms) | Description
    --------------------------------------------------------------------------------
        5e+06 |     215 | Simple curve
        2e+05 |     774 | Curve with markers
        1e+06 |    2411 | Curve with sticks
        1e+04 |    2025 | Error bar curve (vertical bars only)
        1e+04 |     198 | Error bar curve (horizontal and vertical bars)
        1e+06 |     105 | Simple histogram
        1e+03 |     722 | Polar pcolor
        7e+03 |     902 | Simple image
"""

import time

import guidata
import numpy as np
import pytest
from guidata.env import execenv
from guidata.qthelpers import qt_app_context
from guidata.widgets import about
from qtpy import QtWidgets as QW

from plotpy.builder import make
from plotpy.plot import PlotOptions, PlotWindow


class BaseBM:
    """Benchmark object"""

    MAKE_FUNC = make.curve  # to be overriden in subclasses
    WIN_TYPE = "auto"

    @classmethod
    def print_header(cls):
        """Print header for benchmark results"""
        execenv.print(f"PlotPy plot benchmark [{about.get_python_libs_info()}]")
        execenv.print()
        table_header = (
            "N".rjust(10) + " | " + "∆t (ms)".rjust(7) + " | " + "Description"
        ).ljust(80)
        execenv.print(table_header)
        execenv.print("-" * len(table_header))

    def __init__(self, name, nsamples, **options):
        self.name = name
        self.nsamples = int(nsamples)
        self.options = options
        self._item = None

    def compute_data(self):
        raise NotImplementedError

    def make_item(self):
        data = self.compute_data()
        self._item = self.MAKE_FUNC(*data, **self.options)

    def add_to_plot(self, plot):
        assert self._item is not None
        plot.add_item(self._item)

    def start(self):
        # Create plot window
        win = PlotWindow(
            toolbar=True, title=self.name, options=PlotOptions(type=self.WIN_TYPE)
        )
        win.show()
        QW.QApplication.processEvents()
        plot = win.manager.get_plot()

        # Create item
        self.make_item()

        # Benchmarking
        t0 = time.time()
        self.add_to_plot(plot)
        plot.replot()  # Force replot
        QW.QApplication.processEvents()
        dt = (time.time() - t0) * 1e3

        row = f"{self.nsamples:10.0e} | {int(dt):7} | {self.name}"
        execenv.print(row)

        return win


class CurveBM(BaseBM):
    MAKE_FUNC = make.curve
    WIN_TYPE = "curve"

    def compute_data(self):
        x = np.linspace(-10, 10, self.nsamples)
        y = np.sin(np.sin(np.sin(x)))
        return x, y


class HistogramBM(CurveBM):
    MAKE_FUNC = make.histogram

    def compute_data(self):
        data = np.random.normal(size=int(self.nsamples))
        return (data,)


class ErrorBarBM(CurveBM):
    MAKE_FUNC = make.merror

    def __init__(self, name, nsamples, dx=False, **options):
        super().__init__(name, nsamples, **options)
        self.dx = dx

    def compute_data(self):
        x, y = super().compute_data()
        if not self.dx:
            return x, y, x / 100.0
        else:
            return x, y, x / 100.0, x / 20.0


class ImageBM(BaseBM):
    MAKE_FUNC = make.image
    WIN_TYPE = "image"

    def compute_data(self):
        data = np.zeros((int(self.nsamples), int(self.nsamples)), dtype=np.float32)
        m = 10
        step = int(self.nsamples / m)
        for i in range(m):
            for j in range(m):
                data[i * step : (i + 1) * step, j * step : (j + 1) * step] = i * m + j
        return (data,)


class PColorBM(BaseBM):
    MAKE_FUNC = make.pcolor
    WIN_TYPE = "image"

    def compute_data(self):
        N = self.nsamples
        r, th = np.meshgrid(np.linspace(1.0, 16, N), np.linspace(0.0, np.pi, N))
        x = r * np.cos(th)
        y = r * np.sin(th)
        z = 4 * th + r
        return x, y, z


@pytest.mark.skip(reason="Not relevant in automated test suite")
def test_benchmarks():
    """Run benchmark"""
    # Print(informations banner)
    BaseBM.print_header()

    _app = guidata.qapplication()

    # Run benchmarks
    persist = []
    with qt_app_context(exec_loop=True):
        for benchmark in (
            CurveBM("Simple curve", 5e6),
            CurveBM("Curve with markers", 2e5, marker="Ellipse", markersize=10),
            CurveBM("Curve with sticks", 1e6, curvestyle="Sticks"),
            ErrorBarBM("Error bar curve (vertical bars only)", 1e4),
            ErrorBarBM("Error bar curve (horizontal and vertical bars)", 1e4, dx=True),
            HistogramBM("Simple histogram", 1e6, bins=10000),
            PColorBM("Polar pcolor", 1e3),
            ImageBM("Simple image", 7e3, interpolation="antialiasing"),
        ):
            persist.append(benchmark.start())


if __name__ == "__main__":
    test_benchmarks()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Simple application based on plotpy"""

# guitest: show

import sys

import numpy as np
from guidata.configtools import get_icon
from guidata.dataset import (
    ChoiceItem,
    DataSet,
    FloatArrayItem,
    GetAttrProp,
    IntItem,
    StringItem,
    update_dataset,
)
from guidata.dataset.qtwidgets import DataSetEditGroupBox
from guidata.qthelpers import (
    add_actions,
    create_action,
    get_std_icon,
    qt_app_context,
    win32_fix_title_bar_background,
)
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from plotpy import io
from plotpy.builder import make
from plotpy.config import _
from plotpy.plot import PlotOptions, PlotWidget
from plotpy.tests import get_path
from plotpy.widgets import about


class ImageParam(DataSet):
    _hide_data = False
    _hide_size = True
    title = StringItem(_("Title"), default=_("Untitled"))
    data = FloatArrayItem(_("Data")).set_prop("display", hide=GetAttrProp("_hide_data"))
    width = IntItem(
        _("Width"), help=_("Image width (pixels)"), min=1, default=100
    ).set_prop("display", hide=GetAttrProp("_hide_size"))
    height = IntItem(
        _("Height"), help=_("Image height (pixels)"), min=1, default=100
    ).set_prop("display", hide=GetAttrProp("_hide_size"))


class ImageParamNew(ImageParam):
    _hide_data = True
    _hide_size = False
    type = ChoiceItem(_("Type"), (("rand", _("random")), ("zeros", _("zeros"))))


class ImageListWithProperties(QW.QSplitter):
    def __init__(self, parent):
        QW.QSplitter.__init__(self, parent)
        self.imagelist = QW.QListWidget(self)
        self.addWidget(self.imagelist)
        self.properties = DataSetEditGroupBox(_("Properties"), ImageParam)
        self.properties.setEnabled(False)
        self.addWidget(self.properties)


class CentralWidget(QW.QSplitter):
    def __init__(self, parent, toolbar):
        QW.QSplitter.__init__(self, parent)
        self.setContentsMargins(10, 10, 10, 10)
        self.setOrientation(QC.Qt.Vertical)

        imagelistwithproperties = ImageListWithProperties(self)
        self.addWidget(imagelistwithproperties)
        self.imagelist = imagelistwithproperties.imagelist
        self.imagelist.currentRowChanged.connect(self.current_item_changed)
        self.imagelist.itemSelectionChanged.connect(self.selection_changed)
        self.properties = imagelistwithproperties.properties
        self.properties.SIG_APPLY_BUTTON_CLICKED.connect(self.properties_changed)

        self.plot_widget = PlotWidget(
            self,
            options=PlotOptions(type="image", show_contrast=True),
            auto_tools=False,
        )
        self.plot_widget.plot.SIG_LUT_CHANGED.connect(self.lut_range_changed)
        self.item = None  # image item

        self.plot_widget.manager.add_toolbar(toolbar, "default")
        self.plot_widget.register_tools()

        self.addWidget(self.plot_widget)

        self.images = []  # List of ImageParam instances
        self.lut_ranges = []  # List of LUT ranges

        self.setStretchFactor(0, 0)
        self.setStretchFactor(1, 1)
        self.setHandleWidth(10)
        self.setSizes([1, 2])

    def refresh_list(self):
        """Refresh image list"""
        self.imagelist.clear()
        self.imagelist.addItems([image.title for image in self.images])

    def selection_changed(self):
        """Image list: selection changed"""
        row = self.imagelist.currentRow()
        self.properties.setDisabled(row == -1)

    def current_item_changed(self, row):
        """Image list: current image changed"""
        if row == -1:
            return
        image, lut_range = self.images[row], self.lut_ranges[row]
        self.show_data(image.data, lut_range)
        update_dataset(self.properties.dataset, image)
        self.properties.get()

    def lut_range_changed(self):
        """LUT range changed"""
        row = self.imagelist.currentRow()
        self.lut_ranges[row] = self.item.get_lut_range()

    def show_data(self, data, lut_range=None):
        """Show image data"""
        plot = self.plot_widget.plot
        if self.item is not None:
            self.item.set_data(data)
            if lut_range is None:
                lut_range = self.item.get_lut_range()
            self.plot_widget.manager.set_contrast_range(*lut_range)
            self.plot_widget.manager.update_cross_sections()
        else:
            self.item = make.image(data, interpolation="nearest")
            plot.add_item(self.item, z=0)
        plot.select_item(self.item)
        plot.do_autoscale()
        plot.replot()

    def properties_changed(self):
        """The properties 'Apply' button was clicked: updating image"""
        row = self.imagelist.currentRow()
        image = self.images[row]
        update_dataset(image, self.properties.dataset)
        self.refresh_list()
        self.show_data(image.data)

    def add_image(self, image):
        """Add image"""
        self.images.append(image)
        self.lut_ranges.append(None)
        self.refresh_list()
        self.imagelist.setCurrentRow(len(self.images) - 1)
        plot = self.plot_widget.plot
        plot.do_autoscale()

    def add_image_from_file(self, filename):
        """Add image from file"""
        image = ImageParam()
        image.title = str(filename)
        image.data = io.imread(filename, to_grayscale=True)
        image.height, image.width = image.data.shape
        self.add_image(image)

    def remove_image(self, index=None):
        """Remove image"""
        if index is None:
            index = self.imagelist.currentRow()
        del self.images[index]
        del self.lut_ranges[index]
        self.refresh_list()
        if self.imagelist.count() > 0:
            self.imagelist.setCurrentRow(0)
        else:
            self.item = None
            self.plot_widget.plot.del_all_items()
            self.plot_widget.plot.replot()


class MainWindow(QW.QMainWindow):
    """Main Window"""

    def __init__(self):
        super().__init__()
        win32_fix_title_bar_background(self)
        self.setup()

    def setup(self):
        """Setup window parameters"""
        self.setWindowIcon(get_icon("python.png"))
        self.setWindowTitle(_("Application example"))
        self.resize(QC.QSize(600, 800))

        # Welcome message in statusbar:
        status = self.statusBar()
        status.showMessage(_("Welcome to plotpy application example!"), 5000)

        # Set central widget:
        main_toolbar = self.addToolBar("Main")
        toolbar = self.addToolBar("Image")
        self.mainwidget = CentralWidget(self, toolbar)
        self.setCentralWidget(self.mainwidget)

        # File menu
        file_menu = self.menuBar().addMenu(_("File"))
        new_action = create_action(
            self,
            _("New..."),
            shortcut="Ctrl+N",
            icon=get_icon("filenew.png"),
            tip=_("Create a new image"),
            triggered=self.new_image,
        )
        open_action = create_action(
            self,
            _("Open..."),
            shortcut="Ctrl+O",
            icon=get_icon("fileopen.png"),
            tip=_("Open an image"),
            triggered=self.open_image,
        )
        quit_action = create_action(
            self,
            _("Quit"),
            shortcut="Ctrl+Q",
            icon=get_std_icon("DialogCloseButton"),
            tip=_("Quit application"),
            triggered=self.close,
        )
        add_actions(file_menu, (new_action, open_action, None, quit_action))

        # Edit menu
        edit_menu = self.menuBar().addMenu(_("Edit"))
        del_action = create_action(
            self,
            _("Delete"),
            shortcut="Del",
            icon=get_icon("editdelete.png"),
            tip=_("Delete selected image"),
            triggered=self.mainwidget.remove_image,
        )
        add_actions(edit_menu, (del_action,))

        # Help menu
        help_menu = self.menuBar().addMenu("?")
        about_action = create_action(
            self,
            _("About %s...") % "PlotPy",
            icon=get_std_icon("MessageBoxInformation"),
            triggered=about.show_about_dialog,
        )
        add_actions(help_menu, (about_action,))

        add_actions(main_toolbar, (new_action, open_action))

    # ------I/O
    def new_image(self, imagenew=None):
        """Create a new image"""
        if imagenew is None:
            imagenew = ImageParamNew(title=_("Create a new image"))
            if not imagenew.edit(self):
                return
        image = ImageParam()
        image.title = imagenew.title
        if imagenew.type == "zeros":
            image.data = np.zeros((imagenew.width, imagenew.height))
        elif imagenew.type == "rand":
            image.data = np.random.randn(imagenew.width, imagenew.height)
        self.mainwidget.add_image(image)

    def open_image(self, filename=None):
        """Open image file"""
        if filename is None:
            saved_in, saved_out, saved_err = sys.stdin, sys.stdout, sys.stderr
            sys.stdout = None
            filename, _filter = QW.QFileDialog.getOpenFileName(
                self,
                _("Open"),
                "",
                io.iohandler.get_filters("load"),
                "",
                options=QW.QFileDialog.ShowDirsOnly,
            )
            sys.stdin, sys.stdout, sys.stderr = saved_in, saved_out, saved_err
        if filename:
            self.mainwidget.add_image_from_file(filename)


def test_simple_window():
    """Test simple window"""
    with qt_app_context(exec_loop=True):
        window = MainWindow()
        window.show()
        window.new_image(imagenew=ImageParamNew.create(type="rand"))
        window.open_image(filename=get_path("brain.png"))


if __name__ == "__main__":
    test_simple_window()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Testing image show with data update
===================================

The purpose of this test is to implement a simple use case where an image is
updated in a loop. The image is created with random data and the pixel values
are updated in a loop. The test is successful if the image is updated in the
plot window.

It is also a test for the "Update LUT range automatically" feature.
"""

# guitest: show

from __future__ import annotations

import numpy as np
import pytest
from guidata.qthelpers import exec_dialog, qt_app_context
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from plotpy.builder import make
from plotpy.plot import PlotDialog, PlotOptions
from plotpy.tests import data as ptd
from plotpy.tools import LockLUTRangeTool


def get_data(variable_size: bool) -> np.ndarray:
    """Compute 2D Gaussian data and add a narrower Gaussian on top with a random
    position and amplitude."""
    size = np.random.randint(50, 200) if variable_size else 100
    dtype = np.uint16
    amp = np.iinfo(dtype).max * 0.3
    data = ptd.gen_2d_gaussian(size, dtype, sigma=10.0, x0=0.0, y0=0.0, amp=amp)
    # Choose a random position: x0, y0 have to be in the range [-10.0, 10.0]
    x0 = np.random.uniform(-10.0, 10.0)
    y0 = np.random.uniform(-10.0, 10.0)
    # Choose a random amplitude: a has to be in the range [0.1, 0.5]
    a = np.random.uniform(0.1, 0.7) * np.iinfo(dtype).max
    # Add the narrower Gaussian on top
    data += ptd.gen_2d_gaussian(size, dtype, sigma=4.0, x0=x0, y0=y0, amp=a)
    return data


class ImageUpdateDialog(PlotDialog):
    """Dialog box for image update"""

    def __init__(self, title: str, variable_size: bool = False) -> None:
        self.variable_size = variable_size
        options = PlotOptions(title="-", show_contrast=True, type="image")
        super().__init__(title=title, toolbar=True, edit=False, options=options)
        self.resize(600, 600)
        self.timer = QC.QTimer()
        self.item = make.image(get_data(self.variable_size), interpolation="nearest")
        self.item.set_lut_range((15000, 28000))
        plot = self.get_plot()
        plot.add_item(self.item)
        plot.set_active_item(self.item, select=False)
        self.counter = 0
        self.keep_lut_cb: QW.QCheckBox | None = None

    def populate_plot_layout(self) -> None:
        """Populate the plot layout"""
        start_btn = QW.QPushButton("Start image update")
        start_btn.clicked.connect(self.start_image_update)
        self.add_widget(start_btn, 0, 0)
        stop_btn = QW.QPushButton("Stop image update")
        stop_btn.clicked.connect(self.stop_image_update)
        self.add_widget(stop_btn, 0, 1)
        self.keep_lut_cb = QW.QCheckBox()
        self.keep_lut_cb.setChecked(False)
        self.add_widget(self.keep_lut_cb, 0, 2)
        variable_size_cb = QW.QCheckBox("Variable size")
        variable_size_cb.setChecked(self.variable_size)
        variable_size_cb.stateChanged.connect(self.toggle_variable_size)
        self.add_widget(variable_size_cb, 0, 3)
        self.add_widget(self.plot_widget, 1, 0, 1, 0)

    def toggle_variable_size(self, state: int) -> None:
        """Toggle the variable size of the image"""
        self.variable_size = state == QC.Qt.Checked

    def register_tools(self) -> None:
        """Reimplement to connect the Keep LUT range checkbox to the item"""
        mgr = self.get_manager()
        keep_lut_tool = mgr.add_tool(LockLUTRangeTool)
        mgr.add_separator_tool()
        super().register_tools()
        keep_lut_tool.action.toggled.connect(self.keep_lut_cb.setChecked)
        self.keep_lut_cb.setText(keep_lut_tool.action.text())
        self.keep_lut_cb.setToolTip(keep_lut_tool.action.toolTip())
        self.keep_lut_cb.stateChanged.connect(keep_lut_tool.activate)

    def start_image_update(self) -> None:
        """Start updating the image"""
        self.timer.timeout.connect(self.update_image)
        self.timer.start(500)

    def stop_image_update(self) -> None:
        """Stop updating the image"""
        self.timer.stop()

    def update_image(self) -> None:
        """Update the image"""
        data = get_data(self.variable_size)
        self.counter += 1
        plot = self.get_plot()

        # Update the image data
        self.item.set_data(data)
        plot.update_colormap_axis(self.item)
        # Set item as active, again, to force the plot to update the LUT range
        plot.set_active_item(self.item, select=False)

        plot.set_title(f"Image update {self.counter:03d}")
        plot.replot()

    def closeEvent(self, event: QG.QCloseEvent) -> None:
        """Reimplement closeEvent to stop the timer before closing the dialog"""
        self.timer.stop()
        super().closeEvent(event)


@pytest.mark.skip(reason="Not relevant in automated test suite")
def test_image_update() -> None:
    """Testing image update"""
    title = test_image_update.__doc__
    with qt_app_context(exec_loop=False):
        dlg = ImageUpdateDialog(title)
        exec_dialog(dlg)


if __name__ == "__main__":
    test_image_update()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

"""
Scaling functions
-----------------

Overview
^^^^^^^^

The :py:mod:`.scaler` module provides scaling functions for images, thanks to
the C++ scaler engine (`_scaler` extension).

The following functions are available:

* :py:func:`.resize`: resize an image using the scaler engine

Reference
^^^^^^^^^

.. autofunction:: resize
"""

# TODO: Move all _scaler imports in this module and do something to avoid
# the need to import INTERP_LINEAR, INTERP_AA, ... in all modules using the
# scaler (code refactoring between pyplot.imshow,
# styles.BaseImageParam.update_item)

# TODO: Other functions like resize could be written in the future

import numpy as np

from plotpy._scaler import INTERP_AA, INTERP_LINEAR, INTERP_NEAREST, _scale_rect


def resize(data, shape, interpolation=None):
    """Resize array *data* to *shape* (tuple)
    interpolation: 'nearest', 'linear' (default), 'antialiasing'"""
    interpolate = (INTERP_NEAREST,)
    if interpolation is not None:
        interp_dict = {
            "nearest": INTERP_NEAREST,
            "linear": INTERP_LINEAR,
            "antialiasing": INTERP_AA,
        }
        assert interpolation in interp_dict, "invalid interpolation option"
        interp_mode = interp_dict[interpolation]
        if interp_mode in (INTERP_NEAREST, INTERP_LINEAR):
            interpolate = (interp_mode,)
        if interp_mode == INTERP_AA:
            aa = np.ones((5, 5), data.dtype)
            interpolate = (interp_mode, aa)
    out = np.empty(shape)
    src_rect = (0, 0, data.shape[1], data.shape[0])
    dst_rect = (0, 0, out.shape[1], out.shape[0])
    _scale_rect(data, src_rect, out, dst_rect, (1.0, 0.0, None), interpolate)
    return out



from __future__ import annotations

import copy
import sys
from typing import Callable

import matplotlib.colors as pltc
import matplotlib.pyplot as plt
import numpy as np

from plotpy.mathutils.colormap import (
    DEFAULT_COLORMAPS,
    DEFAULT_COLORMAPS_PATH,
    CmapDictType,
    EditableColormap,
    save_colormaps,
)

PERCEPTUALLY_UNIFORM_CMAPS = ["viridis", "plasma", "inferno", "magma", "cividis"]
SEQUENTIAL_CMAPS = [
    "Greys",
    "Purples",
    "Blues",
    "Greens",
    "Oranges",
    "Reds",
    "YlOrBr",
    "YlOrRd",
    "OrRd",
    "PuRd",
    "RdPu",
    "BuPu",
    "GnBu",
    "PuBu",
    "YlGnBu",
    "PuBuGn",
    "BuGn",
    "YlGn",
]
SEQUENTIAL_CMAPS2 = [
    "binary",
    "gist_yarg",
    "gist_gray",
    "gray",
    "bone",
    "pink",
    "spring",
    "summer",
    "autumn",
    "winter",
    "cool",
    "Wistia",
    "hot",
    "afmhot",
    "gist_heat",
    "copper",
]
DIVERGING_CMAPS = [
    "PiYG",
    "PRGn",
    "BrBG",
    "PuOr",
    "RdGy",
    "RdBu",
    "RdYlBu",
    "RdYlGn",
    "Spectral",
    "coolwarm",
    "bwr",
    "seismic",
]
CYCLIC_CMAPS = ["twilight", "twilight_shifted", "hsv"]
QUALITATIVE_CMAPS = [
    "Pastel1",
    "Pastel2",
    "Paired",
    "Accent",
    "Dark2",
    "Set1",
    "Set2",
    "Set3",
    "tab10",
    "tab20",
    "tab20b",
    "tab20c",
]
MISCELLANEOUS_CMAPS = [
    "flag",
    "prism",
    "ocean",
    "gist_earth",
    "terrain",
    "gist_stern",
    "gnuplot",
    "gnuplot2",
    "CMRmap",
    "cubehelix",
    "brg",
    "gist_rainbow",
    "rainbow",
    "jet",
    "turbo",
    "nipy_spectral",
    "gist_ncar",
]

SORTED_MATPLOTLIB_COLORMAPS: list[str] = [
    *PERCEPTUALLY_UNIFORM_CMAPS,
    *SEQUENTIAL_CMAPS,
    *SEQUENTIAL_CMAPS2,
    *DIVERGING_CMAPS,
    *CYCLIC_CMAPS,
    *QUALITATIVE_CMAPS,
    *MISCELLANEOUS_CMAPS,
]


def rgb_colors_to_hex_list(
    colors: list[tuple[int, int, int]],
) -> list[tuple[float, str]]:
    """Convert a list of RGB colors to a list of tuples with the position of the color
    and the color in hex format. Positions evenly distributed between 0 and 1.

    Args:
        colors: list of RGB colors

    Returns:
        list of tuples with the position of the color and the color in hex format
    """
    return [(i / len(colors), pltc.to_hex(color)) for i, color in enumerate(colors)]


def _interpolate(
    val: float, vmin: tuple[float, float, float], vmax: tuple[float, float, float]
):
    """Interpolate between two level of a color.

    Args:
        val: value to interpolate
        vmin: R, G or B tuple from a matplotlib segmented colormap
        vmax: R, G or B tuple from matplotlib segmented colormap

    Returns:
        The interpolated R, G or B component
    """
    interp = (val - vmin[0]) / (vmax[0] - vmin[0])
    return (1 - interp) * vmin[1] + interp * vmax[2]


def std_segmented_cmap_to_hex_list(cmdata: dict[str, list[tuple[float, float, float]]]):
    """Convert a matplotlib segmented colormap to a list of tuples with the position of
    the color and the color in hex format.

    Args:
        cmdata: segmented colormap data

    Returns:
        list of tuples with the position of the color and the color in hex format
    """
    colors: list[tuple[float, str]] = []
    red = np.array(cmdata["red"])
    green = np.array(cmdata["green"])
    blue = np.array(cmdata["blue"])
    indices = sorted(set(red[:, 0]) | set(green[:, 0]) | set(blue[:, 0]))
    for i in indices:
        idxr = red[:, 0].searchsorted(i)
        idxg = green[:, 0].searchsorted(i)
        idxb = blue[:, 0].searchsorted(i)
        compr = _interpolate(i, red[idxr - 1], red[idxr])
        compg = _interpolate(i, green[idxg - 1], green[idxg])
        compb = _interpolate(i, blue[idxb - 1], blue[idxb])
        colors.append((i, pltc.to_hex((compr, compg, compb))))
    return colors


InterpFuncT = Callable[[np.ndarray], np.ndarray]


def func_segmented_cmap_to_hex_list(
    n: int,
    cmap: pltc.LinearSegmentedColormap,
) -> list[tuple[float, str]]:
    """Convert a matplotlib segmented colormap to a list of tuples with the position of
    the color and the color in hex format. The input colormap contains function for each
    color RGB component instead of a list of colors.

    Args:
        n: number of colors to generate
        cmap: segmented colormap

    Returns:
        list of tuples with the position of the color and the color in hex format
    """
    colors = []
    arr = np.linspace(0, 1, n, dtype=float)
    colors = [(i, pltc.to_hex(rgba)) for i, rgba in zip(arr, cmap(arr))]
    return colors


def continuous_to_descrete_cmap(cmap: EditableColormap) -> EditableColormap:
    """Convert a continuous colormap to a descrete one.

    Args:
        cmap: colormap to convert

    Returns:
        descrete colormap
    """
    raw_cmap: tuple[tuple[float, str], ...] = cmap.to_tuples()
    new_raw_cmap: list[tuple[float, str]] = [raw_cmap[0]]
    n = len(raw_cmap)
    coeff = (n - 1) / n
    for i, (pos, color) in enumerate(raw_cmap[1:]):
        prev_pos, prev_color = raw_cmap[i]
        curr_pos, curr_color = pos, color
        new_pos = curr_pos * coeff
        new_raw_cmap.append((new_pos, prev_color))
        new_raw_cmap.append((new_pos, curr_color))
    new_raw_cmap.append(raw_cmap[-1])

    return EditableColormap.from_iterable(new_raw_cmap, name=cmap.name)


def sort_mpl_colormaps(colormaps: CmapDictType) -> CmapDictType:
    """Filter and sort input colormaps to follow the same order (by category) as in the
     matplotlib colormaps documentation. Colormaps not found in the matplotlib
     are filtered out.

    Args:
        colormaps: Dictionnary of colormaps to extract and order

    Returns:
        Filtered and sorted colormaps dictionnary
    """
    ordered_colormaps: CmapDictType = {}
    lower_cmap_names = [cm.lower() for cm in SORTED_MATPLOTLIB_COLORMAPS]
    for lower_name in lower_cmap_names:
        cmap = colormaps.get(lower_name, None)
        if lower_name.endswith("_r"):
            continue
        if cmap is None:
            print(f"Colormap {lower_name} not found in input colormaps.")
            continue
        ordered_colormaps[lower_name] = cmap
    return ordered_colormaps


def append_non_mpl_colormaps(mpl_colormaps: CmapDictType, colormaps: CmapDictType):
    """Append colormaps not found in the matplotlib colormaps to the input colormaps.
     Mutate the input in place.

    Args:
        mpl_colormaps: dictionnary of matplotlib colormaps. Mutated in place.
        colormaps: dictionnary of colormaps to append to the matplotlib colormaps
    """
    colormap_names = set(SORTED_MATPLOTLIB_COLORMAPS)
    for colormap in colormaps.values():
        if colormap.name not in colormap_names:
            print(f"{colormap} not in matplotlib colormaps.")
            mpl_colormaps[colormap.name.lower()] = colormap


def main(cmaps: CmapDictType, out_json_path: str = DEFAULT_COLORMAPS_PATH):
    new_cmaps: dict[str, list[tuple[float, str]]] = {}

    # Uniform colormaps with a .colors attribute that return a list of RGB colors
    cmaps_with_colors = [
        "magma",
        "viridis",
        "inferno",
        "plasma",
        "cividis",
    ]

    # Discrete colormaps, same as uniform colormaps but the colormap must be post
    # processed to become descrete
    descrete_cmaps = [
        "Pastel1",
        "Pastel2",
        "Paired",
        "Accent",
        "Dark2",
        "Set1",
        "Set2",
        "Set3",
    ]

    cmaps_with_colors.extend(descrete_cmaps)

    # Colormaps with a _segmented_data attribute that contains the R, G and B components
    # as lists of tuples
    segmented_cmaps = [
        "coolwarm",
        "bwr",
        "seismic",
    ]

    # Colormaps with a _segmentdata attribute that contains the R, G and B components as
    # functions that return the color for a given position
    interp_cmaps = ["gnuplot2", "CMRmap", "rainbow", "turbo", "afmhot"]

    for cm_name in cmaps_with_colors:
        cmap = plt.get_cmap(cm_name)
        new_cmaps[cm_name] = rgb_colors_to_hex_list(cmap.colors)

    for cm_name in descrete_cmaps:
        cmap = EditableColormap.from_iterable(new_cmaps[cm_name], name=cm_name)
        new_cmaps[cm_name] = list(continuous_to_descrete_cmap(cmap).to_tuples())

    for cm_name in segmented_cmaps:
        cmap = plt.get_cmap(cm_name)
        new_cmaps[cm_name] = std_segmented_cmap_to_hex_list(cmap._segmentdata)

    n = 128

    for cm_name in interp_cmaps:
        cmap = plt.get_cmap(cm_name)
        new_cmaps[cm_name] = func_segmented_cmap_to_hex_list(n, cmap)

    for name, raw_cm in new_cmaps.items():
        cmaps[name.lower()] = EditableColormap.from_iterable(raw_cm, name=name)

    ordered_cmaps = sort_mpl_colormaps(cmaps)
    append_non_mpl_colormaps(ordered_cmaps, cmaps)
    save_colormaps(out_json_path, ordered_cmaps)


if __name__ == "__main__":
    cmaps = copy.deepcopy(DEFAULT_COLORMAPS)
    out_json = sys.argv[1] if len(sys.argv) > 1 else DEFAULT_COLORMAPS_PATH
    main(cmaps, out_json)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Data for tests"""

from __future__ import annotations

import numpy as np


def gen_image1(N: int = 2000, grid: bool = True) -> np.ndarray:
    """Generates a cosine image with a grid, and 4 corners with different values

    Args:
        N: image size
        grid: if True, a grid is added to the image

    Returns:
        image array
    """
    T = np.float32
    x = np.array(np.linspace(-5, 5, N), T)
    img = np.zeros((N, N), T)
    x.shape = (1, N)
    img += x**2
    x.shape = (N, 1)
    img += x**2
    np.cos(img, img)  # inplace cosine
    if not grid:
        return img
    x.shape = (N,)
    for k in range(-5, 5):
        i = x.searchsorted(k)
        if k < 0:
            v = -1.1
        else:
            v = 1.1
        img[i, :] = v
        img[:, i] = v
    m1, m2, m3, m4 = -1.1, -0.3, 0.3, 1.1
    K = 100
    img[:K, :K] = m1  # (0,0)
    img[:K, -K:] = m2  # (0,N)
    img[-K:, -K:] = m3  # (N,N)
    img[-K:, :K] = m4  # (N,0)
    return img


def gen_image2(N: int = 1000, grid: bool = True) -> np.ndarray:
    """Generates a cosine image with a grid, and 4 corners with different values.
    It is a bit different from gen_image1, because the cosine frequency is
    higher and the grid step is smaller.

    Args:
        N: image size
        grid: if True, a grid is added to the image

    Returns:
        image array
    """
    T = np.float64
    TMAX = 32000
    TMIN = 32000
    S = 5.0
    x = np.array(np.linspace(-5 * S, 5 * S, N), T)
    img = np.zeros((N, N), T)
    x.shape = (1, N)
    img += x**2
    x.shape = (N, 1)
    img += x**2
    img = TMAX * np.cos(img / S) + TMIN
    if not grid:
        return img
    x.shape = (N,)
    #    dx = dy = x[1]-x[0]
    for k in range(-5, 5):
        i = x.searchsorted(k)
        if k < 0:
            v = -1.1
        else:
            v = 1.1
        img[i, :] = v
        img[:, i] = v
    m1, m2, m3, m4 = -1.1, -0.3, 0.3, 1.1
    K = 100
    img[:K, :K] = TMAX * m1 + TMIN  # (0,0)
    img[:K, -K:] = TMAX * m2 + TMIN  # (0,N)
    img[-K:, -K:] = TMAX * m3 + TMIN  # (N,N)
    img[-K:, :K] = TMAX * m4 + TMIN  # (N,0)
    return img


def gen_image3(N: int = 1000) -> np.ndarray:
    """Generates a grid image with horizontal and vertical ramps

    Args:
        N: image size

    Returns:
        image array
    """
    NK = 20
    T = float
    img = np.zeros((N, N), T)
    x = np.arange(N, dtype=float)
    x.shape = (1, N)
    DK = N // NK
    for i in range(NK):
        S = i + 1
        y = S * (x // S)
        img[DK * i : DK * (i + 1), :] = y
    return img


def gen_image4(NX: int, NY: int) -> np.ndarray:
    """Generates image data based on a random normal distribution with FFT operations

    Args:
        NX: image size in X
        NY: image size in Y

    Returns:
        image array
    """
    BX, BY = 40, 40
    img = np.random.normal(0, 100, size=(BX, BY))
    timg = np.fft.fftshift(np.fft.fft2(img))
    print(timg.shape)
    cx = NX // 2
    cy = NY // 2
    bx2 = BX // 2
    by2 = BY // 2
    z = np.zeros((NX, NY), np.complex64)
    z[cx - bx2 : cx - bx2 + BX, cy - by2 : cy - by2 + BY] = timg
    z = np.fft.ifftshift(z)
    rev = np.fft.ifft2(z)
    return np.abs(rev)


def gen_xyimage(N: int = 1000) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Generates a cosine image with a grid, and 4 corners with different values

    Args:
        N: image size
        grid: if True, a grid is added to the image

    Returns:
        x, y, data
    """
    N = 1000
    data = gen_image1(N=N)
    x = np.array(np.linspace(-5, 5, N), np.float32)
    data += np.random.normal(0.0, 0.05, size=(N, N))
    return x, (x + 5) ** 0.6, data


def gen_2d_gaussian(size, dtype, x0=0, y0=0, mu=0.0, sigma=2.0, amp=None) -> np.ndarray:
    """Creating 2D Gaussian (-10 <= x <= 10 and -10 <= y <= 10)"""
    xydata = np.linspace(-10, 10, size)
    x, y = np.meshgrid(xydata, xydata)
    if amp is None:
        amp = np.iinfo(dtype).max * 0.5
    t = (np.sqrt((x - x0) ** 2 + (y - y0) ** 2) - mu) ** 2
    return np.array(amp * np.exp(-t / (2.0 * sigma**2)), dtype=dtype)


def gen_1d_gaussian(
    size, x0=0, mu=0.0, sigma=2.0, amp=None
) -> tuple[np.ndarray, np.ndarray]:
    """Creating 1D Gaussian (-10 <= x <= 10)"""
    x = np.linspace(-10, 10, size)
    if amp is None:
        amp = 1.0
    t = (np.abs(x - x0) - mu) ** 2
    y = np.array(amp * np.exp(-t / (2.0 * sigma**2)), dtype=float)
    return x, y


def gen_xyz_data() -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Create a X, Y, Z data set for contour detection features"""
    delta = 0.025
    x, y = np.arange(-3.0, 3.0, delta), np.arange(-2.0, 2.0, delta)
    X, Y = np.meshgrid(x, y)
    Z1 = np.exp(-(X**2) - Y**2)
    Z2 = np.exp(-((X - 1) ** 2) - (Y - 1) ** 2)
    Z = (Z1 - Z2) * 2
    return X, Y, Z



# -*- coding: utf-8 -*-

# pylint: disable=unused-import
# flake8: noqa

from .annotation import (
    AnnotatedCircle,
    AnnotatedEllipse,
    AnnotatedObliqueRectangle,
    AnnotatedPoint,
    AnnotatedRectangle,
    AnnotatedPolygon,
    AnnotatedSegment,
    AnnotatedXRange,
    AnnotatedYRange,
    AnnotatedShape,
)
from .contour import ContourItem, create_contour_items
from .curve import CurveItem, ErrorBarCurveItem
from .grid import GridItem
from .histogram import HistogramItem
from .image import (
    BaseImageItem,
    Histogram2DItem,
    ImageFilterItem,
    ImageItem,
    MaskedImageItem,
    MaskedXYImageItem,
    QuadGridItem,
    RawImageItem,
    RGBImageItem,
    TrImageItem,
    XYImageFilterItem,
    XYImageItem,
    assemble_imageitems,
    compute_trimageitems_original_size,
    get_image_from_plot,
    get_image_from_qrect,
    get_image_in_shape,
    get_items_in_rectangle,
    get_plot_qrect,
)
from .image.masked import MaskedArea, MaskedImageItem, MaskedXYImageItem
from .label import (
    AbstractLabelItem,
    DataInfoLabel,
    LabelItem,
    LegendBoxItem,
    ObjectInfo,
    RangeComputation,
    XRangeComputation,
    YRangeComputation,
    RangeComputation2d,
    RangeInfo,
    SelectedLegendBoxItem,
)
from .polygonmap import PolygonMapItem
from .shape import (
    AbstractShape,
    Axes,
    CircleSVGShape,
    EllipseShape,
    Marker,
    ObliqueRectangleShape,
    PointShape,
    PolygonShape,
    RectangleShape,
    RectangleSVGShape,
    SegmentShape,
    SquareSVGShape,
    XRangeSelection,
    YRangeSelection,
)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Curve plotting test with high DPI"""

# guitest: show,skip

import os

import numpy as np
import pytest
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import vistools as ptv


@pytest.mark.skip(reason="This test is not relevant for the automated test suite")
def test_plot_highdpi():
    """Curve plotting test with high DPI"""
    # When setting the QT_SCALE_FACTOR to "2", performance is degraded, due to the
    # increased number of points to be drawn. As a workaround, we use the downsampling
    # feature to reduce the number of points to be drawn.
    # We also enable the antialiasing feature to improve the rendering of the curves,
    # which is also affected by the high DPI setting.
    # See https://github.com/PlotPyStack/PlotPy/issues/10
    os.environ["QT_SCALE_FACTOR"] = "2"

    npoints = 5000000  # 5M points are needed to see the difference
    dsamp_factor = npoints // 50000  # Max 50000 points to be drawn
    x = np.linspace(-10, 10, npoints)
    y = np.sin(np.sin(np.sin(x)))
    with qt_app_context(exec_loop=True):
        _win = ptv.show_items(
            [
                make.curve(x, y, color="b", dsamp_factor=dsamp_factor, use_dsamp=True),
                make.legend("TR"),
            ],
            curve_antialiasing=True,
            wintitle=test_plot_highdpi.__doc__,
            title="Curves with high DPI",
            plot_type="curve",
        )


if __name__ == "__main__":
    test_plot_highdpi()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Image superposition test"""

# guitest: show

import numpy as np
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import get_path
from plotpy.tools import EllipseTool, PlaceAxesTool, PolygonTool, RectangleTool


def create_window():
    gridparam = make.gridparam(
        background="black", minor_enabled=(False, False), major_style=(":", "gray", 1)
    )
    win = make.dialog(
        toolbar=True,
        wintitle="Image superposition test",
        gridparam=gridparam,
        type="image",
        size=(800, 600),
    )
    for toolklass in (RectangleTool, EllipseTool, PolygonTool, PlaceAxesTool):
        win.manager.add_tool(toolklass)
    return win


def test_imagesuperp():
    """Test image superposition"""
    filename = get_path("brain.png")
    with qt_app_context(exec_loop=True):
        win = create_window()
        image1 = make.image(filename=filename, title="Original", colormap="gray")
        data2 = np.array(image1.data[:, :150], copy=True)
        data2[:100] = data2[-100:] = 0
        image2 = make.image(data2, title="Modified", alpha_function="step")
        plot = win.manager.get_plot()
        plot.add_item(image1, z=0)
        plot.add_item(image2, z=1)
        plot.set_items_readonly(False)
        image2.set_readonly(True)
        win.manager.get_itemlist_panel().show()
        win.show()


if __name__ == "__main__":
    test_imagesuperp()



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

import numpy as np
from guidata.dataset import (
    BeginGroup,
    BoolItem,
    ChoiceItem,
    ColorItem,
    DataSet,
    EndGroup,
    FloatItem,
    GetAttrProp,
    ImageChoiceItem,
    IntItem,
    StringItem,
)
from qtpy import QtGui as QG

from plotpy._scaler import INTERP_AA, INTERP_LINEAR, INTERP_NEAREST
from plotpy.config import _
from plotpy.constants import LUTAlpha
from plotpy.mathutils.colormap import (
    ALL_COLORMAPS,
    LARGE_ICON_HEIGHT,
    LARGE_ICON_ORIENTATION,
    LARGE_ICON_WIDTH,
    build_icon_from_cmap_name,
)
from plotpy.styles.base import ItemParameters

if TYPE_CHECKING:
    from guidata.dataset import DataSet

    from plotpy.items import (
        BaseImageItem,
        ImageFilterItem,
        ImageItem,
        MaskedImageItem,
        MaskedXYImageItem,
        RawImageItem,
        RGBImageItem,
        TrImageItem,
    )
    from plotpy.plot import BasePlot


def _create_choices(
    dataset: DataSet, item: ImageChoiceItem, value: Any
) -> list[tuple[str, str, Callable[[str], QG.QIcon]]]:
    """Create the list of choices for the colormap item."""
    choices: list[tuple[str, str, Callable[[str], QG.QIcon]]] = []
    for cmap in ALL_COLORMAPS.values():
        choices.append(
            (
                cmap.name,
                cmap.name,
                lambda name: build_icon_from_cmap_name(
                    name,
                    LARGE_ICON_WIDTH,
                    LARGE_ICON_HEIGHT,
                    LARGE_ICON_ORIENTATION,
                    1,
                ),
            )
        )
    return choices


# TODO: Use an "enum" like LUTAlpha for the interpolation mode as well
# (and eventually for other parameters)


class BaseImageParam(DataSet):
    """Base class for image parameters."""

    _multiselection = False
    label = StringItem(_("Image title"), default=_("Image")).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    alpha_function = ChoiceItem(
        _("Alpha function"),
        LUTAlpha.get_choices(),
        default=LUTAlpha.NONE.value,
        help=LUTAlpha.get_help(),
    )
    alpha = FloatItem(
        _("Global alpha"), default=1.0, min=0, max=1, help=_("Global alpha value")
    )
    _hide_colormap = False
    colormap = (
        ImageChoiceItem(_("Colormap"), _create_choices, default="jet")
        .set_prop("display", hide=GetAttrProp("_hide_colormap"))
        .set_prop("display", size=(LARGE_ICON_WIDTH, LARGE_ICON_HEIGHT))
    )
    invert_colormap = BoolItem(_("Invert colormap"), default=False)

    keep_lut_range = BoolItem(
        _("Lock LUT range when updating data"),
        default=False,
        help=_(
            "If enabled, the LUT range is not updated when the image data changes."
            "<br>This allows to keep the same color scale for different successive "
            "images. <br><br>"
            "<u>Note:</u> It has no effect when a new image is added to the plot."
        ),
    )

    interpolation = ChoiceItem(
        _("Interpolation"),
        [
            (0, _("None (nearest pixel)")),
            (1, _("Linear interpolation")),
            (2, _("2x2 antialiasing filter")),
            (3, _("3x3 antialiasing filter")),
            (5, _("5x5 antialiasing filter")),
        ],
        default=0,
        help=_("Image interpolation type"),
    )

    _formats = BeginGroup(_("Statistics string formatting"))
    xformat = StringItem(_("X-Axis"), default=r"%.1f")
    yformat = StringItem(_("Y-Axis"), default=r"%.1f")
    zformat = StringItem(_("Z-Axis"), default=r"%.1f")
    _end_formats = EndGroup(_("Statistics string formatting"))

    def get_units(self) -> tuple[str, str, str]:
        """Get the units for the x, y and z axes.

        Returns:
            The units for the x, y and z axes.
        """
        try:
            xunit = self.xformat.split()[1]
        except IndexError:
            xunit = ""
        try:
            yunit = self.yformat.split()[1]
        except IndexError:
            yunit = ""
        try:
            zunit = self.zformat.split()[1]
        except IndexError:
            zunit = ""
        return xunit, yunit, zunit

    def update_param(self, item: BaseImageItem) -> None:
        """Update the parameters from the given image item.

        Args:
            item: The image item to update the parameters from.
        """
        self.label = str(item.title().text())
        cmap = item.get_color_map()
        if cmap is not None:
            self.colormap = cmap.name
            self.invert_colormap = cmap.invert
        interpolation = item.get_interpolation()
        mode = interpolation[0]

        if mode == INTERP_NEAREST:
            self.interpolation = 0
        elif mode == INTERP_LINEAR:
            self.interpolation = 1
        else:
            size = interpolation[1].shape[0]
            self.interpolation = size

    def update_item(self, item: BaseImageItem) -> None:
        """Update the given image item from the parameters.

        Args:
            item: The image item to update.
        """
        if isinstance(self.alpha_function, LUTAlpha):
            self.alpha_function = self.alpha_function.value
        plot: BasePlot = item.plot()
        if plot is not None:
            plot.blockSignals(True)  # Avoid unwanted calls of update_param
            # triggered by the setter methods below
        item.setTitle(self.label)
        item.set_color_map(self.colormap, self.invert_colormap)
        size = self.interpolation

        if size == 0:
            mode = INTERP_NEAREST
        elif size == 1:
            mode = INTERP_LINEAR
        else:
            mode = INTERP_AA
        item.set_interpolation(mode, size)
        if plot is not None:
            plot.blockSignals(False)


class QuadGridParam(DataSet):
    """Parameters for the grid of a quad item."""

    _multiselection = False
    label = StringItem(_("Image title"), default=_("Image")).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    alpha_function = ChoiceItem(
        _("Alpha function"),
        LUTAlpha.get_choices(),
        default=LUTAlpha.NONE.value,
        help=LUTAlpha.get_help(),
    )
    alpha = FloatItem(
        _("Global alpha"), default=1.0, min=0, max=1, help=_("Global alpha value")
    )
    _hide_colormap = False
    colormap = ImageChoiceItem(_("Colormap"), _create_choices, default="jet").set_prop(
        "display", hide=GetAttrProp("_hide_colormap")
    )
    invert_colormap = BoolItem(_("Invert colormap"), default=False)

    interpolation = ChoiceItem(
        _("Interpolation"),
        [(0, _("Quadrangle interpolation")), (1, _("Flat"))],
        default=0,
        help=_(
            "Image interpolation type, Flat mode use fixed u,v interpolation parameters"
        ),
    )
    uflat = FloatItem(
        _("Fixed U interpolation parameter"),
        default=0.5,
        min=0.0,
        max=1.0,
        help=_("For flat mode only"),
    )
    vflat = FloatItem(
        _("Fixed V interpolation parameter"),
        default=0.5,
        min=0.0,
        max=1.0,
        help=_("For flat mode only"),
    )
    grid = BoolItem(_("Show grid"), default=False)
    gridcolor = ColorItem(_("Grid lines color"), default="black")

    def update_param(self, item: BaseImageItem) -> None:
        """Update the parameters from the given image item.

        Args:
            item: The image item to update the parameters from.
        """
        self.label = str(item.title().text())
        cmap = item.get_color_map()
        if cmap is not None:
            self.colormap = cmap.name
            self.invert_colormap = cmap.invert
        interp, uflat, vflat = item.interpolate
        self.interpolation = interp
        self.uflat = uflat
        self.vflat = vflat
        self.grid = item.grid

    def update_item(self, item: BaseImageItem) -> None:
        """Update the given image item from the parameters.

        Args:
            item: The image item to update.
        """
        if isinstance(self.alpha_function, LUTAlpha):
            self.alpha_function = self.alpha_function.value
        plot = item.plot()
        if plot is not None:
            plot.blockSignals(True)  # Avoid unwanted calls of update_param
            # triggered by the setter methods below
        item.setTitle(self.label)
        item.set_color_map(self.colormap, self.invert_colormap)
        item.interpolate = (self.interpolation, self.uflat, self.vflat)
        item.grid = self.grid
        if plot is not None:
            plot.blockSignals(False)


class RawImageParam(BaseImageParam):
    """Parameters for a raw image."""

    _hide_background = False
    background = ColorItem(_("Background color"), default="#000000").set_prop(
        "display", hide=GetAttrProp("_hide_background")
    )

    def update_param(self, item: RawImageItem) -> None:
        """Update the parameters from the given image item.

        Args:
            item: The image item from which to update the parameters.
        """
        super().update_param(item)
        self.background = str(QG.QColor(item.bg_qcolor).name())

    def update_item(self, item: RawImageItem) -> None:
        """Update the given image item from the parameters.

        Args:
            item: The image item to update.
        """
        super().update_item(item)
        plot: BasePlot = item.plot()
        if plot is not None:
            plot.blockSignals(True)  # Avoid unwanted calls of update_param
            # triggered by the setter methods below
        item.set_background_color(self.background)
        if plot is not None:
            plot.blockSignals(False)


class RawImageParam_MS(RawImageParam):
    """Parameters for a raw image (multiselection)."""

    _multiselection = True


ItemParameters.register_multiselection(RawImageParam, RawImageParam_MS)


class XYImageParam(RawImageParam):
    """Parameters for an XY image."""

    pass


class XYImageParam_MS(XYImageParam):
    """Parameters for an XY image (multiselection)."""

    _multiselection = True


ItemParameters.register_multiselection(XYImageParam, XYImageParam_MS)


class ImageParam(RawImageParam):
    """Parameters for an image."""

    lock_position = BoolItem(
        _("Lock position"),
        _("Position"),
        default=True,
        help=_("Locked images are not movable with the mouse"),
    )
    _xdata = BeginGroup(_("Image placement along X-axis"))
    xmin = FloatItem(_("x|min"), default=None, check=False)
    xmax = FloatItem(_("x|max"), default=None, check=False)
    _end_xdata = EndGroup(_("Image placement along X-axis"))
    _ydata = BeginGroup(_("Image placement along Y-axis"))
    ymin = FloatItem(_("y|min"), default=None, check=False)
    ymax = FloatItem(_("y|max"), default=None, check=False)
    _end_ydata = EndGroup(_("Image placement along Y-axis"))

    def update_param(self, item: ImageItem) -> None:
        """Update the parameters from the given image item.

        Args:
            item: The image item from which to update the parameters.
        """
        super().update_param(item)
        self.xmin = item.xmin
        self.ymin = item.ymin
        self.xmax = item.xmax
        self.ymax = item.ymax

    def update_item(self, item: ImageItem) -> None:
        """Update the given image item from the parameters.

        Args:
            item: The image item to update.
        """
        super().update_item(item)
        plot: BasePlot = item.plot()
        if plot is not None:
            plot.blockSignals(True)  # Avoid unwanted calls of update_param
            # triggered by the setter methods below
        item.xmin = self.xmin
        item.xmax = self.xmax
        item.ymin = self.ymin
        item.ymax = self.ymax
        item.update_bounds()
        item.update_border()
        if plot is not None:
            plot.blockSignals(False)


class ImageParam_MS(ImageParam):
    """Parameters for an image (multiselection)."""

    _multiselection = True


ItemParameters.register_multiselection(ImageParam, ImageParam_MS)


class RGBImageParam(ImageParam):
    """Parameters for an RGB image."""

    _hide_background = True
    _hide_colormap = True

    def update_item(self, item: RGBImageItem) -> None:
        """Update the given image item from the parameters.

        Args:
            item: The image item to update.
        """
        super().update_item(item)
        plot = item.plot()
        if plot is not None:
            plot.blockSignals(True)  # Avoid unwanted calls of update_param
            # triggered by the setter methods below
        item.recompute_alpha_channel()
        if plot is not None:
            plot.blockSignals(False)


class RGBImageParam_MS(RGBImageParam):
    """Parameters for an RGB image (multiselection)."""

    _multiselection = True


ItemParameters.register_multiselection(RGBImageParam, RGBImageParam_MS)


class MaskedImageParamMixin(DataSet):
    """Mixin for masked image parameters."""

    g_mask = BeginGroup(_("Mask"))
    filling_value = FloatItem(_("Filling value"), allow_none=True)
    show_mask = BoolItem(_("Show image mask"), default=False)
    alpha_masked = FloatItem(_("Masked area alpha"), default=0.7, min=0, max=1)
    alpha_unmasked = FloatItem(_("Unmasked area alpha"), default=0.0, min=0, max=1)
    _g_mask = EndGroup(_("Mask"))

    def _update_item(self, item: MaskedImageItem) -> None:
        """Update the given masked image item from the parameters.

        Args:
            item: The masked image item to update.
        """
        plot: BasePlot = item.plot()
        if plot is not None:
            plot.blockSignals(True)  # Avoid unwanted calls of update_param
            # triggered by the setter methods below
        item.update_mask()
        if plot is not None:
            plot.blockSignals(False)


class MaskedImageParam(ImageParam, MaskedImageParamMixin):
    """Parameters for a masked image."""

    def update_item(self, item: MaskedImageItem) -> None:
        """Update the given image item from the parameters.

        Args:
            item: The image item to update.
        """
        super().update_item(item)
        self._update_item(item)


class MaskedXYImageParam(XYImageParam, MaskedImageParamMixin):
    """Parameters for a masked XY image."""

    def update_item(self, item: MaskedXYImageItem) -> None:
        """Update the given image item from the parameters.

        Args:
            item: The image item to update.
        """
        super().update_item(item)
        self._update_item(item)


class MaskedImageParam_MS(MaskedImageParam):
    """Parameters for a masked image (multiselection)."""

    _multiselection = True


class MaskedXYImageParam_MS(MaskedXYImageParam):
    """Parameters for a masked XY image (multiselection)."""

    _multiselection = True


ItemParameters.register_multiselection(MaskedImageParam, MaskedImageParam_MS)
ItemParameters.register_multiselection(MaskedXYImageParam, MaskedXYImageParam_MS)


class ImageFilterParam(BaseImageParam):
    """Parameters for an image filter."""

    label = StringItem(_("Title"), default=_("Filter"))
    g1 = BeginGroup(_("Bounds"))
    xmin = FloatItem(_("x|min"), default=0.0)
    xmax = FloatItem(_("x|max"), default=1.0)
    ymin = FloatItem(_("y|min"), default=0.0)
    ymax = FloatItem(_("y|max"), default=1.0)
    _g1 = EndGroup("sub-group")
    use_source_cmap = BoolItem(
        _("Use image colormap and level"), _("Color map"), default=True
    )

    def update_param(self, item: ImageFilterItem) -> None:
        """Update the parameters from the given image filter item.

        Args:
            item: The image filter item from which to update the parameters.
        """
        self.xmin, self.ymin, self.xmax, self.ymax = item.border_rect.get_rect()
        self.use_source_cmap = item.use_source_cmap
        super().update_param(item)

    def update_imagefilter(self, item: ImageFilterItem) -> None:
        """Update the given image filter item from the parameters.

        Args:
            item: The image filter item to update.
        """
        m, M = item.get_lut_range()
        set_range = False
        if not self.use_source_cmap and item.use_source_cmap:
            set_range = True
        item.use_source_cmap = self.use_source_cmap
        if set_range:
            item.set_lut_range([m, M])
        self.update_item(item)
        item.border_rect.set_rect(self.xmin, self.ymin, self.xmax, self.ymax)


class TrImageParam(RawImageParam):
    """Parameters for a transformed image."""

    lock_position = BoolItem(
        _("Lock position"),
        _("Position"),
        default=False,
        help=_("Locked images are not movable with the mouse"),
    )

    _crop = BeginGroup(_("Crop")).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    crop_left = IntItem(_("Left"), default=0)
    crop_right = IntItem(_("Right"), default=0)
    crop_top = IntItem(_("Top"), default=0)
    crop_bottom = IntItem(_("Bottom"), default=0)
    _end_crop = EndGroup(_("Cropping")).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    _ps = BeginGroup(_("Pixel size")).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    dx = FloatItem(_("Width (dx)"), default=1.0)
    dy = FloatItem(_("Height (dy)"), default=1.0)
    _end_ps = EndGroup(_("Pixel size")).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    _pos = BeginGroup(_("Translate, rotate and flip"))
    pos_x0 = FloatItem(_("x<sub>CENTER</sub>"), default=0.0).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    hflip = BoolItem(_("Flip horizontally"), default=False).set_prop("display", col=1)
    pos_y0 = FloatItem(_("y<sub>CENTER</sub>"), default=0.0).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    vflip = BoolItem(_("Flip vertically"), default=False).set_prop("display", col=1)
    pos_angle = FloatItem(_("θ (°)"), default=0.0).set_prop("display", col=0)
    _end_pos = EndGroup(_("Translate, rotate and flip"))

    def update_param(self, item: TrImageItem) -> None:
        """Update the parameters from the given image item.

        Args:
            item: The image item from which to update the parameters.
        """
        super().update_param(item)
        # we don't get crop info from the image because
        # its not easy to extract from the transform
        # and TrImageItem keeps it's crop information
        # directly in this DataSet

    def update_item(self, item: TrImageItem) -> None:
        """Update the given image item from the parameters.

        Args:
            item: The image item to update.
        """
        RawImageParam.update_item(self, item)
        plot: BasePlot = item.plot()
        if plot is not None:
            plot.blockSignals(True)  # Avoid unwanted calls of update_param
            # triggered by the setter methods below
        item.set_transform(*self.get_transform())
        if plot is not None:
            plot.blockSignals(False)

    def get_transform(self) -> tuple[float, float, float, float, float, bool, bool]:
        """Get the transformation parameters.

        Returns:
            The transformation parameters: x0, y0, angle, dx, dy, hflip, vflip.
        """
        return (
            self.pos_x0,
            self.pos_y0,
            self.pos_angle * np.pi / 180,
            self.dx,
            self.dy,
            self.hflip,
            self.vflip,
        )

    def set_transform(
        self,
        x0: float,
        y0: float,
        angle: float,
        dx: float = 1.0,
        dy: float = 1.0,
        hflip: bool = False,
        vflip: bool = False,
    ) -> None:
        """Set the transformation parameters.

        Args:
            x0: The x-coordinate of the center of the image.
            y0: The y-coordinate of the center of the image.
            angle: The rotation angle in degrees.
            dx: The width of the image.
            dy: The height of the image.
            hflip: Flip the image horizontally.
            vflip: Flip the image vertically.
        """
        self.pos_x0 = x0
        self.pos_y0 = y0
        self.pos_angle = angle * 180 / np.pi
        self.dx = dx
        self.dy = dy
        self.hflip = hflip
        self.vflip = vflip

    def set_crop(self, left: int, top: int, right: int, bottom: int) -> None:
        """Set the crop parameters.

        Args:
            left: The left crop.
            top: The top crop.
            right: The right crop.
            bottom: The bottom crop.
        """
        self.crop_left = left
        self.crop_right = right
        self.crop_top = top
        self.crop_bottom = bottom

    def get_crop(self) -> tuple[int, int, int, int]:
        """Get the crop parameters.

        Returns:
            The crop parameters: left, top, right, bottom.
        """
        return (self.crop_left, self.crop_top, self.crop_right, self.crop_bottom)


class TrImageParam_MS(TrImageParam):
    """Parameters for a transformed image (multiselection)."""

    _multiselection = True


ItemParameters.register_multiselection(TrImageParam, TrImageParam_MS)



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import TYPE_CHECKING

import guidata.io
import numpy as np
import numpy.ma as ma
from qtpy import QtCore as QC

from plotpy import io
from plotpy._scaler import INTERP_NEAREST, _scale_rect, _scale_xy
from plotpy.config import _
from plotpy.interfaces import (
    IBaseImageItem,
    IBasePlotItem,
    IExportROIImageItemType,
    IHistDataSource,
    IVoiImageItemType,
)
from plotpy.items.image.standard import ImageItem, XYImageItem
from plotpy.styles.image import MaskedImageParam, MaskedXYImageParam

if TYPE_CHECKING:
    import guidata.io
    import qwt.scale_map
    from qtpy.QtCore import QRectF
    from qtpy.QtGui import QPainter

    from plotpy.plot import BasePlot


class MaskedArea:
    """Defines masked areas for a masked image item

    Args:
        geometry: Geometry of the area ('rectangular' or anything else for circular)
        x0: X coordinate of first point
        y0: Y coordinate of first point
        x1: X coordinate of second point
        y1: Y coordinate of second point
        inside: True if the area is inside the geometry, False if it is outside
    """

    def __init__(
        self,
        geometry: str | None = None,
        x0: float | None = None,
        y0: float | None = None,
        x1: float | None = None,
        y1: float | None = None,
        inside: bool | None = None,
    ) -> None:
        self.geometry = geometry
        self.x0 = x0
        self.y0 = y0
        self.x1 = x1
        self.y1 = y1
        self.inside = inside

    def __eq__(self, other):
        return (
            self.geometry == other.geometry
            and self.x0 == other.x0
            and self.y0 == other.y0
            and self.x1 == other.x1
            and self.y1 == other.y1
            and self.inside == other.inside
        )

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        for name in ("geometry", "inside", "x0", "y0", "x1", "y1"):
            writer.write(getattr(self, name), name)

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        self.geometry = reader.read("geometry")
        self.inside = reader.read("inside")
        for name in ("x0", "y0", "x1", "y1"):
            setattr(self, name, reader.read(name, func=reader.read_float))


class MaskedImageMixin:
    """Masked image mixin

    This class is a mixin providing array mask features to image items. It is used by
    :py:class:`.MaskedImageItem` and :py:class:`.MaskedXYImageItem` classes.

    Args:
        mask: 2D masked array
    """

    def __init__(self, mask: ma.MaskedArray | None = None) -> None:
        self.orig_data: np.ndarray | None = None
        self._mask: ma.MaskedArray | None = mask
        self._mask_filename: str | None = None
        self._masked_areas: list[MaskedArea] = []
        # ImageItem and XYImageItem attributes:
        self.data: ma.MaskedArray | None = None
        self.param: MaskedImageParam | MaskedXYImageParam | None = None

    # ---- Pickle methods -------------------------------------------------------
    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        writer.write(self.get_mask_filename(), group_name="mask_fname")
        writer.write_object_list(self._masked_areas, "masked_areas")

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        mask_fname = reader.read(group_name="mask_fname", func=reader.read_unicode)
        masked_areas = reader.read_object_list("masked_areas", MaskedArea)
        if mask_fname:
            self.set_mask_filename(mask_fname)
            self.load_mask_data()
        elif masked_areas and self.data is not None:
            self.set_masked_areas(masked_areas)
            self.apply_masked_areas()

    # ---- Public API -----------------------------------------------------------
    def update_mask(self) -> None:
        """Update mask"""
        if isinstance(self.data, np.ma.MaskedArray):
            filling_value = self.param.filling_value
            if filling_value is None or (
                isinstance(filling_value, float) and np.isnan(filling_value)
            ):
                # Use a safe default for integer types when filling_value is None or NaN
                if np.issubdtype(self.data.dtype, np.integer):
                    val = np.array(0, dtype=self.data.dtype)
                else:
                    val = np.array(np.nan, dtype=self.data.dtype)
            else:
                # If filling_value is out of bounds for data dtype, get the default
                # fill value for that dtype
                if np.issubdtype(self.data.dtype, np.integer):
                    info = np.iinfo(self.data.dtype)
                    if filling_value < info.min or filling_value > info.max:
                        # Using the default fill value thanks to the get_fill_value()
                        # method of MaskedArray:
                        filling_value = self.data.get_fill_value()
                val = np.array(filling_value).astype(self.data.dtype)
            self.data.set_fill_value(val)

    def set_mask(self, mask: ma.MaskedArray) -> None:
        """Set image mask

        Args:
            mask: 2D masked array
        """
        self.data.mask = mask

    def get_mask(self) -> ma.MaskedArray | None:
        """Get image mask

        Returns:
            2D masked array
        """
        return self.data.mask

    def set_mask_filename(self, fname: str) -> None:
        """Set mask filename

        Args:
            fname: Mask filename

        There are two ways for pickling mask data of `MaskedImageItem` objects:

            1. using the mask filename (as for data itself)
            2. using the mask areas (`MaskedArea` instance, see set_mask_areas)

        When saving objects, the first method is tried and then, if no
        filename has been defined for mask data, the second method is used.
        """
        self._mask_filename = fname

    def get_mask_filename(self) -> str:
        """Get mask filename

        Returns:
            Mask filename
        """
        return self._mask_filename

    def load_mask_data(self) -> None:
        """Load mask data from file"""
        data = io.imread(self.get_mask_filename(), to_grayscale=True)
        self.set_mask(data)
        self._mask_changed()

    def set_masked_areas(self, areas: list[MaskedArea]) -> None:
        """Set masked areas

        Args:
            areas: List of masked areas

        .. seealso:: :py:meth:`.set_mask_filename`
        """
        self._masked_areas = areas

    def get_masked_areas(self) -> list[MaskedArea]:
        """Get masked areas

        Returns:
            List of masked areas
        """
        return self._masked_areas

    def add_masked_area(
        self, geometry: str, x0: float, y0: float, x1: float, y1: float, inside: bool
    ) -> None:
        """Add masked area

        Args:
            geometry: Area geometry
            x0: top left x coordinate
            y0: top left y coordinate
            x1: bottom right x coordinate
            y1: bottom right y coordinate
            inside: Inside or outside
        """
        area = MaskedArea(geometry=geometry, x0=x0, y0=y0, x1=x1, y1=y1, inside=inside)
        for _area in self._masked_areas:
            if area == _area:
                return
        self._masked_areas.append(area)

    def _mask_changed(self) -> None:
        """Emit the :py:data:`.baseplot.BasePlot.SIG_MASK_CHANGED` signal"""
        plot: BasePlot = self.plot()
        if plot is not None:
            plot.SIG_MASK_CHANGED.emit(self)

    def apply_masked_areas(self) -> None:
        """Apply masked areas"""
        for area in self._masked_areas:
            if area.geometry == "rectangular":
                self.mask_rectangular_area(
                    area.x0,
                    area.y0,
                    area.x1,
                    area.y1,
                    area.inside,
                    trace=False,
                    do_signal=False,
                )
            else:
                self.mask_circular_area(
                    area.x0,
                    area.y0,
                    area.x1,
                    area.y1,
                    area.inside,
                    trace=False,
                    do_signal=False,
                )
        self._mask_changed()

    def mask_all(self) -> None:
        """Mask all pixels"""
        self.data.mask = True
        self._mask_changed()

    def unmask_all(self) -> None:
        """Unmask all pixels"""
        self.data.mask = np.ma.nomask
        self.set_masked_areas([])
        self._mask_changed()

    def mask_rectangular_area(
        self,
        x0: float,
        y0: float,
        x1: float,
        y1: float,
        inside: bool = True,
        trace: bool = True,
        do_signal: bool = True,
    ) -> None:
        """Mask rectangular area

        Args:
            x0: top left x coordinate
            y0: top left y coordinate
            x1: bottom right x coordinate
            y1: bottom right y coordinate
            inside: if True (default), mask the inside of the area, otherwise
             mask the outside
            trace: if True (default), add the area to the list of masked areas
            do_signal: if True (default), emit the
             :py:data:`.baseplot.BasePlot.SIG_MASK_CHANGED` signal
        """
        ix0, iy0, ix1, iy1 = self.get_closest_index_rect(x0, y0, x1, y1)
        if inside:
            self.data[iy0:iy1, ix0:ix1] = np.ma.masked
        else:
            indexes = np.ones(self.data.shape, dtype=bool)
            indexes[iy0:iy1, ix0:ix1] = False
            self.data[indexes] = np.ma.masked
        if trace:
            self.add_masked_area("rectangular", x0, y0, x1, y1, inside)
        if do_signal:
            self._mask_changed()

    def mask_circular_area(
        self,
        x0: float,
        y0: float,
        x1: float,
        y1: float,
        inside: bool = True,
        trace: bool = True,
        do_signal: bool = True,
    ) -> None:
        """Mask circular area (as a circle inscribed in a rectangle)

        Args:
            x0: top left x coordinate
            y0: top left y coordinate
            x1: bottom right x coordinate
            y1: bottom right y coordinate
            inside: if True (default), mask the inside of the area, otherwise
             mask the outside
            trace: if True (default), add the area to the list of masked areas
            do_signal: if True (default), emit the
             :py:data:`.baseplot.BasePlot.SIG_MASK_CHANGED` signal
        """
        ix0, iy0, ix1, iy1 = self.get_closest_index_rect(x0, y0, x1, y1)
        xc, yc = 0.5 * (x0 + x1), 0.5 * (y0 + y1)
        radius = 0.5 * (x1 - x0)
        xdata, ydata = self.get_x_values(ix0, ix1), self.get_y_values(iy0, iy1)

        for ix in range(ix0, ix1):
            for iy in range(iy0, iy1):
                distance = np.sqrt(
                    (xdata[ix - ix0] - xc) ** 2 + (ydata[iy - iy0] - yc) ** 2
                )
                if inside:
                    if distance <= radius:
                        self.data[iy, ix] = np.ma.masked
                elif distance > radius:
                    self.data[iy, ix] = np.ma.masked

        if not inside:
            self.mask_rectangular_area(x0, y0, x1, y1, inside, trace=False)
        if trace:
            self.add_masked_area("circular", x0, y0, x1, y1, inside)
        if do_signal:
            self._mask_changed()

    def is_mask_visible(self) -> bool:
        """Return mask visibility

        Returns:
            True if mask is visible, False otherwise
        """
        return self.param.show_mask

    def set_mask_visible(self, state: bool) -> None:
        """Set mask visibility

        Args:
            state: True to show mask, False to hide it
        """
        self.param.show_mask = state
        plot = self.plot()
        if plot is not None:
            plot.replot()

    def _set_data(self, data: np.ndarray) -> None:
        """Set image data

        Args:
            data: 2D NumPy array
        """
        self.orig_data = data
        self.data = data.view(np.ma.MaskedArray)
        self.set_mask(self._mask)
        self._mask = None  # removing reference to this temporary array
        if self.param.filling_value is None:
            self.param.filling_value = self.data.get_fill_value()
        self.update_mask()


class MaskedImageItem(ImageItem, MaskedImageMixin):
    """Masked image item

    Args:
        data: 2D NumPy array
        mask: 2D NumPy array
        param: image parameters
    """

    __implements__ = (
        IBasePlotItem,
        IBaseImageItem,
        IHistDataSource,
        IVoiImageItemType,
        IExportROIImageItemType,
    )

    def __init__(
        self,
        data: np.ndarray | None = None,
        mask: np.ndarray | None = None,
        param: MaskedImageParam | None = None,
    ) -> None:
        self.orig_data = None
        MaskedImageMixin.__init__(self, mask=mask)
        ImageItem.__init__(self, data=data, param=param)

    # ---- BaseImageItem API ---------------------------------------------------
    def get_default_param(self) -> MaskedImageParam:
        """Return instance of the default MaskedImageParam DataSet"""
        return MaskedImageParam(_("Image"))

    # ---- Pickle methods -------------------------------------------------------
    def __reduce__(self) -> tuple:
        """Reduce object to pickled state"""
        fname = self.get_filename()
        if fname is None:
            fn_or_data = self.data
        else:
            fn_or_data = fname
        state = (
            self.param,
            self.get_lut_range(),
            fn_or_data,
            self.z(),
            self.get_mask_filename(),
            self.get_masked_areas(),
        )
        res = (self.__class__, (), state)
        return res

    def __setstate__(self, state: tuple) -> None:
        """Set object state from pickled state"""
        param, lut_range, fn_or_data, z, mask_fname, old_masked_areas = state
        if old_masked_areas and isinstance(old_masked_areas[0], MaskedArea):
            masked_areas = old_masked_areas
        else:
            # Compatibility with old format
            masked_areas = []
            for geometry, x0, y0, x1, y1, inside in old_masked_areas:
                area = MaskedArea(
                    geometry=geometry, x0=x0, y0=y0, x1=x1, y1=y1, inside=inside
                )
                masked_areas.append(area)
        self.param = param
        if isinstance(fn_or_data, str):
            self.set_filename(fn_or_data)
            self.load_data(lut_range)
        elif fn_or_data is not None:  # should happen only with previous API
            self.set_data(fn_or_data, lut_range=lut_range)
        self.setZ(z)
        self.param.update_item(self)
        if mask_fname is not None:
            self.set_mask_filename(mask_fname)
            self.load_mask_data()
        elif masked_areas and self.data is not None:
            self.set_masked_areas(masked_areas)
            self.apply_masked_areas()

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        ImageItem.serialize(self, writer)
        MaskedImageMixin.serialize(self, writer)

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        ImageItem.deserialize(self, reader)
        MaskedImageMixin.deserialize(self, reader)

    # ---- BaseImageItem API ----------------------------------------------------
    def draw_image(
        self,
        painter: QPainter,
        canvasRect: QRectF,
        src_rect: tuple[float, float, float, float],
        dst_rect: tuple[float, float, float, float],
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
    ) -> None:
        """Draw image

        Args:
            painter: Painter
            canvasRect: Canvas rectangle
            src_rect: Source rectangle
            dst_rect: Destination rectangle
            xMap: X axis scale map
            yMap: Y axis scale map
        """
        ImageItem.draw_image(self, painter, canvasRect, src_rect, dst_rect, xMap, yMap)
        if self.data is None:
            return
        if self.is_mask_visible():
            _a, _b, bg, _cmap = self.lut
            # pylint: disable=too-many-function-args
            alpha_masked = (
                np.uint32(255 * self.param.alpha_masked + 0.5).clip(0, 255) << 24
            )
            alpha_unmasked = (
                np.uint32(255 * self.param.alpha_unmasked + 0.5).clip(0, 255) << 24
            )
            cmap = np.array(
                [
                    np.uint32(0x000000 & 0xFFFFFF) | alpha_unmasked,
                    np.uint32(0xFFFFFF & 0xFFFFFF) | alpha_masked,
                ],
                dtype=np.uint32,
            )
            lut = (1, 0, bg, cmap)
            shown_data = np.ma.getmaskarray(self.data)
            src2 = self._rescale_src_rect(src_rect)
            dst_rect = tuple([int(i) for i in dst_rect])
            dest = _scale_rect(
                shown_data, src2, self._offscreen, dst_rect, lut, (INTERP_NEAREST,)
            )
            qrect = QC.QRectF(
                QC.QPointF(dest[0], dest[1]), QC.QPointF(dest[2], dest[3])
            )
            painter.drawImage(qrect, self._image, qrect)

    # ---- RawImageItem API -----------------------------------------------------
    def set_data(
        self, data: np.ndarray, lut_range: list[float, float] | None = None
    ) -> None:
        """Set image data

        Args:
            data: 2D NumPy array
            lut_range: LUT range -- tuple (levelmin, levelmax) (Default value = None)
        """
        super().set_data(data, lut_range)
        MaskedImageMixin._set_data(self, data)


class MaskedXYImageItem(XYImageItem, MaskedImageMixin):
    """XY masked image item

    Args:
        x: 1D NumPy array, must be increasing
        y: 1D NumPy array, must be increasing
        data: 2D NumPy array
        mask: 2D NumPy array
        param: image parameters
    """

    __implements__ = (
        IBasePlotItem,
        IBaseImageItem,
        IHistDataSource,
        IVoiImageItemType,
        IExportROIImageItemType,
    )

    def __init__(
        self,
        x: np.ndarray | None = None,
        y: np.ndarray | None = None,
        data: np.ndarray | None = None,
        mask: np.ndarray | None = None,
        param: MaskedXYImageParam | None = None,
    ) -> None:
        self.orig_data = None
        MaskedImageMixin.__init__(self, mask=mask)
        XYImageItem.__init__(self, x=x, y=y, data=data, param=param)

    # ---- BaseImageItem API ---------------------------------------------------
    def get_default_param(self) -> MaskedXYImageParam:
        """Return instance of the default MaskedXYImageParam DataSet"""
        return MaskedXYImageParam(_("Image"))

    # ---- Pickle methods -------------------------------------------------------
    def __reduce__(self) -> tuple:
        """Reduce object to pickled state"""
        fname = self.get_filename()
        if fname is None:
            fn_or_data = self.data
        else:
            fn_or_data = fname
        state = (
            self.param,
            self.get_lut_range(),
            fn_or_data,
            self.x,
            self.y,
            self.z(),
            self.get_mask_filename(),
            self.get_masked_areas(),
        )
        res = (self.__class__, (), state)
        return res

    def __setstate__(self, state: tuple) -> None:
        """Set object state from pickled state"""
        param, lut_range, fn_or_data, x, y, z, mask_fname, old_masked_areas = state
        if old_masked_areas and isinstance(old_masked_areas[0], MaskedArea):
            masked_areas = old_masked_areas
        else:
            # Compatibility with old format
            masked_areas = []
            for geometry, x0, y0, x1, y1, inside in old_masked_areas:
                area = MaskedArea(
                    geometry=geometry, x0=x0, y0=y0, x1=x1, y1=y1, inside=inside
                )
                masked_areas.append(area)
        self.param = param
        if isinstance(fn_or_data, str):
            self.set_filename(fn_or_data)
            self.load_data(lut_range)
        elif fn_or_data is not None:  # should happen only with previous API
            self.set_data(fn_or_data, lut_range=lut_range)
        self.set_xy(x, y)
        self.setZ(z)
        self.param.update_item(self)
        if mask_fname is not None:
            self.set_mask_filename(mask_fname)
            self.load_mask_data()
        elif masked_areas and self.data is not None:
            self.set_masked_areas(masked_areas)
            self.apply_masked_areas()

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        XYImageItem.serialize(self, writer)
        MaskedImageMixin.serialize(self, writer)

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        XYImageItem.deserialize(self, reader)
        MaskedImageMixin.deserialize(self, reader)

    # ---- BaseImageItem API ----------------------------------------------------
    def draw_image(
        self,
        painter: QPainter,
        canvasRect: QRectF,
        src_rect: tuple[float, float, float, float],
        dst_rect: tuple[float, float, float, float],
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
    ) -> None:
        """Draw image

        Args:
            painter: Painter
            canvasRect: Canvas rectangle
            src_rect: Source rectangle
            dst_rect: Destination rectangle
            xMap: X axis scale map
            yMap: Y axis scale map
        """
        XYImageItem.draw_image(
            self, painter, canvasRect, src_rect, dst_rect, xMap, yMap
        )
        if self.data is None:
            return
        if self.is_mask_visible():
            _a, _b, bg, _cmap = self.lut
            # pylint: disable=too-many-function-args
            alpha_masked = (
                np.uint32(255 * self.param.alpha_masked + 0.5).clip(0, 255) << 24
            )
            alpha_unmasked = (
                np.uint32(255 * self.param.alpha_unmasked + 0.5).clip(0, 255) << 24
            )
            cmap = np.array(
                [
                    np.uint32(0x000000 & 0xFFFFFF) | alpha_unmasked,
                    np.uint32(0xFFFFFF & 0xFFFFFF) | alpha_masked,
                ],
                dtype=np.uint32,
            )
            lut = (1, 0, bg, cmap)
            shown_data = np.ma.getmaskarray(self.data)

            xytr = self.x, self.y, src_rect
            dst_rect = tuple([int(i) for i in dst_rect])
            dest = _scale_xy(
                shown_data, xytr, self._offscreen, dst_rect, lut, (INTERP_NEAREST,)
            )
            qrect = QC.QRectF(
                QC.QPointF(dest[0], dest[1]), QC.QPointF(dest[2], dest[3])
            )
            painter.drawImage(qrect, self._image, qrect)

    def set_data(
        self, data: np.ndarray, lut_range: list[float, float] | None = None
    ) -> None:
        """Set image data

        Args:
            data: 2D NumPy array
            lut_range: LUT range -- tuple (levelmin, levelmax) (Default value = None)
        """
        super().set_data(data, lut_range)
        MaskedImageMixin._set_data(self, data)



# -*- coding: utf-8 -*-

# pylint: disable=unused-import
# flake8: noqa

from .base import BasePlot, BasePlotOptions
from .manager import PlotManager
from .plotwidget import (
    PlotDialog,
    PlotOptions,
    PlotWidget,
    PlotWindow,
    SubplotWidget,
    SyncPlotWindow,
    SyncPlotDialog,
    set_widget_title_icon,
)



# -*- coding: utf-8 -*-
#
# For licensing and distribution details, please read carefully xgrid/__init__.py

"""
Get rectangular selection from image with SVG shape
"""

# guitest: show

from guidata.env import execenv
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import get_path
from plotpy.tests.data import gen_image4
from plotpy.tests.tools.test_get_segment import SEG_AXES_COORDS, PatchedSelectDialog
from plotpy.tools import RectangularShapeTool
from plotpy.widgets.selectdialog import select_with_shape_tool


class SVGToolExample(RectangularShapeTool):
    """Tool to select a rectangular area and create a pattern from it"""

    TITLE = "Pattern selection tool"
    ICON = "your_icon.svg"
    AVOID_NULL_SHAPE = True
    SVG_FNAME = get_path("svg_tool.svg")

    def create_shape(self):
        """Create shape to be drawn"""
        with open(self.SVG_FNAME, "rb") as svg_file:
            svg_data = svg_file.read()
        shape = make.svg("rectangle", svg_data, 0, 0, 1, 1, "SVG")
        self.set_shape_style(shape)
        return shape, 0, 2


def test_get_rectangle_with_svg():
    """Test get_rectangle_with_svg"""
    with qt_app_context():
        image = make.image(data=gen_image4(200, 200), colormap="gray")
        shape = select_with_shape_tool(
            None, SVGToolExample, image, "Test", tooldialogclass=PatchedSelectDialog
        )
        if shape is not None:
            rect = shape.get_rect()
            if execenv.unattended:
                assert [round(i) for i in list(rect)] == SEG_AXES_COORDS
            elif rect is not None:
                print("Area:", rect)


if __name__ == "__main__":
    test_get_rectangle_with_svg()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
EditPointTool test

This plotpy tool can be used to edit (move and/or add points) to a curve.
"""

# guitest: show

from __future__ import annotations

import numpy as np
from guidata.qthelpers import exec_dialog, qt_app_context
from qtpy import QtWidgets as QW

from plotpy.builder import make
from plotpy.config import _
from plotpy.interfaces.items import ICurveItemType
from plotpy.plot.plotwidget import PlotDialog
from plotpy.tools import EditPointTool


def callback_function(tool: EditPointTool):
    """Callback function to be called by the tool after a point has moved

    Args:
        tool: EditPointTool instance that will use this callback
    """
    print("Indexed changes:", tool.get_changes())


def make_new_bbox(dialog: PlotDialog):
    """Add a new button to the dialog to insert a point at the current selection

    Args:
        dialog: PlotDialog instance
    """
    if (
        dialog.button_box is not None
        and dialog.button_layout is not None
        and (tool := dialog.manager.get_tool(EditPointTool)) is not None
    ):
        insert_btn = QW.QPushButton(_("Insert point"), dialog)
        insert_btn.clicked.connect(tool.trigger_insert_point_at_selection)
        dialog.button_layout.insertWidget(0, insert_btn)


def edit_curve(
    cdata: tuple[tuple[float, float], ...],
) -> tuple[tuple[float, float], ...]:
    """
    Plot curves and return selected point(s) coordinates

    Args:
        cdata: tuple of curves to plot

    Returns:
        tuple of modified curves
    """
    win: PlotDialog = make.dialog(
        wintitle=_("Select one point then press OK to accept"),
        edit=True,
        type="curve",
        size=(800, 600),
    )
    default = win.manager.add_tool(
        EditPointTool,
        title="Test",
        end_callback=callback_function,
    )
    default.activate()
    plot = win.manager.get_plot()
    for cx, cy in cdata[:-1]:
        item = make.mcurve(cx, cy)
        plot.add_item(item)
    item = make.mcurve(*cdata[-1], "r-+")
    plot.add_item(item)
    plot.set_active_item(item)
    plot.unselect_item(item)
    make_new_bbox(win)
    exec_dialog(win)
    return [curve.get_data() for curve in plot.get_items(item_type=ICurveItemType)]


def test_edit_curve() -> None:
    """Test"""
    with qt_app_context():
        x = np.linspace(-10, 10, num=100)
        y = 0.25 * np.sin(np.sin(np.sin(x * 0.5)))
        x2 = np.linspace(-10, 10, num=100)
        y2 = np.sin(np.sin(np.sin(x2)))
        new_cdata = edit_curve(((x, y), (x2, y2), (x, np.sin(2 * y))))
        edit_curve(new_cdata)
        print((y2 == new_cdata[1][1]).all())


if __name__ == "__main__":
    test_edit_curve()



# -*- coding: utf-8 -*-

from __future__ import annotations

import math
import sys
from typing import TYPE_CHECKING

import numpy as np
from guidata.dataset import update_dataset
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qwt import QwtSymbol

from plotpy.config import CONF, _
from plotpy.coords import canvas_to_axes
from plotpy.interfaces import IBasePlotItem, ISerializableType, IShapeItemType
from plotpy.items.shape.base import AbstractShape
from plotpy.styles.shape import ShapeParam

if TYPE_CHECKING:
    import guidata.io
    import qwt.scale_map
    import qwt.symbol
    from qtpy.QtCore import QPointF, QRectF
    from qtpy.QtGui import QBrush, QPainter, QPen, QPolygonF

    from plotpy.interfaces import IItemType
    from plotpy.plot.base import BasePlot
    from plotpy.styles.base import ItemParameters


class PolygonShape(AbstractShape):
    """Polygon shape class

    Args:
        points: List of point coordinates
        closed: True if the polygon is closed, False otherwise
        shapeparam: Shape parameters
    """

    __implements__ = (IBasePlotItem, ISerializableType)
    ADDITIONNAL_POINTS = 0  # Number of points which are not part of the shape
    LINK_ADDITIONNAL_POINTS = False  # Link additionnal points with dotted lines
    CLOSED = True
    _icon_name = "polygon.png"

    def __init__(
        self,
        points: list[tuple[float, float]] | np.ndarray | None = None,
        closed: bool | None = None,
        shapeparam: ShapeParam | None = None,
    ) -> None:
        super().__init__()
        self.closed = self.CLOSED if closed is None else closed
        self.selected = False

        if shapeparam is None:
            self.shapeparam = ShapeParam(_("Shape"), icon="rectangle.png")
        else:
            self.shapeparam = shapeparam
            self.shapeparam.update_item(self)

        self.pen = QG.QPen()
        self.brush = QG.QBrush()
        self.symbol = QwtSymbol.NoSymbol
        self.sel_pen = QG.QPen()
        self.sel_brush = QG.QBrush()
        self.sel_symbol = QwtSymbol.NoSymbol
        self.points: np.ndarray | None = None
        self.set_points(points)

    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (IShapeItemType, ISerializableType)

    def __reduce__(self) -> tuple[type, tuple, tuple]:
        """Return state information for pickling"""
        self.shapeparam.update_param(self)
        state = (self.shapeparam, self.points, self.closed, self.z())
        return (PolygonShape, (), state)

    def __setstate__(self, state: tuple) -> None:
        """Set state information for unpickling"""
        self.shapeparam, self.points, self.closed, z = state
        self.setZ(z)
        self.shapeparam.update_item(self)

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        self.shapeparam.update_param(self)
        writer.write(self.shapeparam, group_name="shapeparam")
        writer.write(self.points, group_name="points")
        writer.write(self.closed, group_name="closed")
        writer.write(self.z(), group_name="z")

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        self.closed = reader.read("closed")
        self.shapeparam = ShapeParam(_("Shape"), icon="rectangle.png")
        reader.read("shapeparam", instance=self.shapeparam)
        self.shapeparam.update_item(self)
        self.points = reader.read(group_name="points", func=reader.read_array)
        self.setZ(reader.read("z"))

    # ----Public API-------------------------------------------------------------

    def set_style(self, section: str, option: str) -> None:
        """Set style for this item

        Args:
            section: Section
            option: Option
        """
        self.shapeparam.read_config(CONF, section, option)
        self.shapeparam.update_item(self)

    def set_points(self, points: list[tuple[float, float]] | np.ndarray | None) -> None:
        """Set polygon points

        Args:
            points: List of point coordinates
        """
        if points is None:
            self.points = np.zeros((0, 2), float)
        else:
            self.points = np.array(points, float)
            assert self.points.shape[1] == 2

    def get_points(self) -> np.ndarray:
        """Return polygon points

        Returns:
            Polygon points (array of shape (N, 2))
        """
        return self.points

    def set_closed(self, state: bool) -> None:
        """Set closed state

        Args:
            state: True if the polygon is closed, False otherwise
        """
        self.closed = state

    def is_closed(self) -> bool:
        """Return True if the polygon is closed, False otherwise

        Returns:
            True if the polygon is closed, False otherwise
        """
        return self.closed

    def get_center(self) -> tuple[float, float]:
        """Return the center of the polygon

        Returns:
            Center of the polygon
        """
        if self.points is not None and self.points.size > 0:
            return self.points.mean(axis=0)
        return 0.0, 0.0

    def boundingRect(self) -> QC.QRectF:
        """Return the bounding rectangle of the data

        Returns:
            Bounding rectangle of the data
        """
        poly = QG.QPolygonF()
        if self.ADDITIONNAL_POINTS:
            shape_points = self.points[: -self.ADDITIONNAL_POINTS]
        else:
            shape_points = self.points
        for i in range(shape_points.shape[0]):
            poly.append(QC.QPointF(shape_points[i, 0], shape_points[i, 1]))
        return poly.boundingRect()

    def is_empty(self) -> bool:
        """Return True if the item is empty

        Returns:
            True if the item is empty, False otherwise
        """
        return len(self.points) == 0

    def get_bounding_rect_coords(self) -> tuple[float, float, float, float]:
        """Return bounding rectangle coordinates (in plot coordinates)

        Returns:
            Bounding rectangle coordinates (in plot coordinates)
        """
        poly = QG.QPolygonF()
        shape_points = self.points[: -self.ADDITIONNAL_POINTS]
        for i in range(shape_points.shape[0]):
            poly.append(QC.QPointF(shape_points[i, 0], shape_points[i, 1]))
        return poly.boundingRect().getCoords()

    def transform_points(
        self, xMap: qwt.scale_map.QwtScaleMap, yMap: qwt.scale_map.QwtScaleMap
    ) -> QPolygonF:
        """Transform points to canvas coordinates

        Args:
            xMap: X axis scale map
            yMap: Y axis scale map

        Returns:
            Transformed points
        """
        points = QG.QPolygonF()
        for i in range(self.points.shape[0]):
            points.append(
                QC.QPointF(
                    xMap.transform(self.points[i, 0]), yMap.transform(self.points[i, 1])
                )
            )
        return points

    def get_reference_point(self) -> tuple[float, float] | None:
        """Return a reference point for the item

        Returns:
            Reference point for the item
        """
        if self.points.size:
            return self.points.mean(axis=0)

    def get_pen_brush(
        self, xMap: qwt.scale_map.QwtScaleMap, yMap: qwt.scale_map.QwtScaleMap
    ) -> tuple[QPen, QBrush, qwt.symbol.QwtSymbol]:
        """Get pen, brush and symbol for the item

        Args:
            xMap: X axis scale map
            yMap: Y axis scale map

        Returns:
            Tuple with pen, brush and symbol for the item
        """
        if self.selected:
            pen = self.sel_pen
            brush = self.sel_brush
            sym = self.sel_symbol
        else:
            pen = self.pen
            brush = self.brush
            sym = self.symbol
        if self.points.size > 0:
            x0, y0 = self.get_reference_point()
            xx0 = xMap.transform(x0)
            yy0 = yMap.transform(y0)
            try:
                # Optimized version in PyQt >= v4.5
                t0 = QG.QTransform.fromTranslate(xx0, yy0)
            except AttributeError:
                # Fallback for PyQt <= v4.4
                t0 = QG.QTransform().translate(xx0, yy0)
            tr = brush.transform()
            tr = tr * t0
            brush = QG.QBrush(brush)
            brush.setTransform(tr)
        return pen, brush, sym

    def draw(
        self,
        painter: QPainter,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QRectF,
    ) -> None:
        """Draw the item

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle
        """
        pen, brush, symbol = self.get_pen_brush(xMap, yMap)
        painter.setRenderHint(QG.QPainter.Antialiasing)
        painter.setPen(pen)
        painter.setBrush(brush)
        points = self.transform_points(xMap, yMap)
        if self.ADDITIONNAL_POINTS:
            # Slice indexing is not supported by PySide6, so we convert the `QPolygonF`
            # object to a list before converting it back to a `QPolygonF` object:
            shape_points = QG.QPolygonF(list(points)[: -self.ADDITIONNAL_POINTS])
            other_points = QG.QPolygonF(list(points)[-self.ADDITIONNAL_POINTS :])
        else:
            shape_points = points
            other_points = []
        if self.closed:
            painter.drawPolygon(shape_points)
        else:
            painter.drawPolyline(shape_points)
        if symbol != QwtSymbol.NoSymbol:
            symbol.drawSymbols(painter, points)
        if self.LINK_ADDITIONNAL_POINTS and other_points:
            pen2 = painter.pen()
            pen2.setStyle(QC.Qt.DotLine)
            painter.setPen(pen2)
            painter.drawPolyline(other_points)

    def poly_hit_test(self, plot: BasePlot, ax: int, ay: int, pos: QPointF) -> tuple:
        """Return a tuple (distance, attach point, inside, other_object)

        Args:
            plot: Plot
            ax: X axis index
            ay: Y axis index
            pos: Position

        Returns:
            Tuple with four elements (distance, attach point, inside, other_object).
        """
        pos = QC.QPointF(pos)
        dist = sys.maxsize
        handle = -1
        Cx, Cy = pos.x(), pos.y()
        poly = QG.QPolygonF()
        pts = self.points
        for i in range(pts.shape[0]):
            # Compute distance to the line segment in canvas coordinates
            px = plot.transform(ax, pts[i, 0])
            py = plot.transform(ay, pts[i, 1])
            if i < pts.shape[0] - self.ADDITIONNAL_POINTS:
                poly.append(QC.QPointF(px, py))
            d = (Cx - px) ** 2 + (Cy - py) ** 2
            if d < dist:
                dist = d
                handle = i
        inside = poly.containsPoint(QC.QPointF(Cx, Cy), QC.Qt.OddEvenFill)
        return math.sqrt(dist), handle, inside, None

    def hit_test(self, pos: QPointF) -> tuple[float, float, bool, None]:
        """Return a tuple (distance, attach point, inside, other_object)

        Args:
            pos: Position

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).

        Description of the returned values:

        * distance: distance in pixels (canvas coordinates) to the closest
           attach point
        * attach point: handle of the attach point
        * inside: True if the mouse button has been clicked inside the object
        * other_object: if not None, reference of the object which will be
           considered as hit instead of self
        """
        if not self.plot():
            return sys.maxsize, 0, False, None
        return self.poly_hit_test(self.plot(), self.xAxis(), self.yAxis(), pos)

    def add_local_point(self, pos: tuple[float, float]) -> int:
        """Add a point in canvas coordinates (local coordinates)

        Args:
            pos: Position

        Returns:
            Handle of the added point
        """
        pt = canvas_to_axes(self, pos)
        return self.add_point(pt)

    def add_point(self, pt: tuple[float, float]) -> int:
        """Add a point in axis coordinates

        Args:
            pt: Position

        Returns:
            Handle of the added point
        """
        if self.points.size == 0:
            self.points = np.array([pt])
            return 0
        N, _ = self.points.shape
        self.points = np.resize(self.points, (N + 1, 2))
        self.points[N, :] = pt
        return N

    def del_point(self, handle: int) -> int:
        """Delete a point

        Args:
            handle: Handle

        Returns:
            Handle of the deleted point
        """
        self.points = np.delete(self.points, handle, 0)
        if handle < len(self.points):
            return handle
        else:
            return self.points.shape[0] - 1

    def move_point_to(
        self, handle: int, pos: tuple[float, float], ctrl: bool = False
    ) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        self.points[handle, :] = pos

    def move_shape(self, old_pos: QC.QPointF, new_pos: QC.QPointF) -> None:
        """Translate the shape such that old_pos becomes new_pos in axis coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        dx = new_pos[0] - old_pos[0]
        dy = new_pos[1] - old_pos[1]
        self.points += np.array([[dx, dy]])

    def update_item_parameters(self) -> None:
        """Update item parameters (dataset) from object properties"""
        self.shapeparam.update_param(self)

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """
        self.update_item_parameters()
        itemparams.add("ShapeParam", self, self.shapeparam)

    def set_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Change the appearance of this item according
        to the parameter set provided

        Args:
            itemparams: Item parameters
        """
        update_dataset(self.shapeparam, itemparams.get("ShapeParam"), visible_only=True)
        self.shapeparam.update_item(self)


assert_interfaces_valid(PolygonShape)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Test for MaskedImageItem dtype changes

This test verifies that changing the data type of a masked image from float to
integer types (like float64 to uint8) doesn't raise RuntimeWarnings when the
filling_value is NaN or None.
"""

from __future__ import annotations

import warnings

import numpy as np

from plotpy.builder import make


def test_masked_image_dtype_change_from_float_to_uint8():
    """Test changing masked image dtype from float64 to uint8 with NaN filling_value"""
    # Create a float64 masked array with NaN as fill value (the default)
    data = np.random.rand(10, 10).astype(np.float64)
    mask = np.zeros_like(data, dtype=bool)
    mask[3:5, 3:5] = True
    masked_data = np.ma.masked_array(data, mask=mask)

    # Create masked image item
    item = make.maskedimage(masked_data, colormap="viridis", show_mask=True)

    # Verify initial state
    assert item.data.dtype == np.float64

    # Set filling_value to NaN (typical for float images)
    item.param.filling_value = np.nan

    # Change data dtype to uint8
    new_data = (masked_data * 255).astype(np.uint8)
    item.set_data(new_data)

    # This should not raise a RuntimeWarning
    with warnings.catch_warnings(record=True) as warning_list:
        warnings.simplefilter("always")
        item.update_mask()

    # Check that no RuntimeWarning was raised
    runtime_warnings = [
        w for w in warning_list if issubclass(w.category, RuntimeWarning)
    ]
    assert len(runtime_warnings) == 0, "RuntimeWarning should not be raised"

    # Verify the data type changed
    assert item.data.dtype == np.uint8


def test_masked_image_dtype_change_with_none_filling_value():
    """Test changing masked image dtype with None filling_value"""
    # Create a float64 masked array
    data = np.random.rand(10, 10).astype(np.float64)
    mask = np.zeros_like(data, dtype=bool)
    mask[3:5, 3:5] = True
    masked_data = np.ma.masked_array(data, mask=mask)

    # Create masked image item
    item = make.maskedimage(masked_data, colormap="viridis", show_mask=True)

    # Set filling_value to None
    item.param.filling_value = None

    # Change data dtype to uint8
    new_data = (masked_data * 255).astype(np.uint8)
    item.set_data(new_data)

    # This should not raise a RuntimeWarning
    with warnings.catch_warnings(record=True) as warning_list:
        warnings.simplefilter("always")
        item.update_mask()

    # Check that no RuntimeWarning was raised
    runtime_warnings = [
        w for w in warning_list if issubclass(w.category, RuntimeWarning)
    ]
    assert len(runtime_warnings) == 0, "RuntimeWarning should not be raised"


def test_masked_image_filling_value_defaults():
    """Test that filling_value defaults are appropriate for different dtypes"""
    test_dtypes = [np.uint8, np.uint16, np.int16, np.float32, np.float64]

    for dtype in test_dtypes:
        data = np.random.rand(5, 5)
        if np.issubdtype(dtype, np.integer):
            data = (data * 100).astype(dtype)
        else:
            data = data.astype(dtype)

        mask = np.zeros_like(data, dtype=bool)
        mask[2:3, 2:3] = True
        masked_data = np.ma.masked_array(data, mask=mask)

        item = make.maskedimage(masked_data, colormap="viridis", show_mask=True)

        # Set filling_value to NaN (typical initial state for float images)
        item.param.filling_value = np.nan

        # This should handle the conversion gracefully
        with warnings.catch_warnings(record=True) as warning_list:
            warnings.simplefilter("always")
            item.update_mask()

        # Check that no RuntimeWarning was raised
        runtime_warnings = [
            w for w in warning_list if issubclass(w.category, RuntimeWarning)
        ]
        assert len(runtime_warnings) == 0, f"RuntimeWarning for dtype {dtype}"


if __name__ == "__main__":
    test_masked_image_dtype_change_from_float_to_uint8()
    test_masked_image_dtype_change_with_none_filling_value()
    test_masked_image_filling_value_defaults()
    print("All tests passed!")



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import TYPE_CHECKING

from guidata.dataset import (
    BoolItem,
    DataSet,
    FloatItem,
    GetAttrProp,
    ImageChoiceItem,
    IntItem,
    StringItem,
)
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qwt import QwtPlotCurve

from plotpy.config import _
from plotpy.styles.base import (
    CURVESTYLE_CHOICES,
    CURVESTYLE_NAME,
    ItemParameters,
    LineStyleItem,
    SymbolItem,
)

if TYPE_CHECKING:
    from plotpy.items import CurveItem


class CurveParam(DataSet):
    """Dataset defining the parameters of a CurveItem"""

    _multiselection = False
    label = StringItem(_("Title"), default="").set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    line = LineStyleItem(_("Line"))
    symbol = SymbolItem(_("Symbol"))
    shade = FloatItem(_("Shadow"), default=0.0, min=0.0, max=1.0)
    curvestyle = ImageChoiceItem(_("Curve style"), CURVESTYLE_CHOICES, default="Lines")
    baseline = FloatItem(_("Baseline"), default=0.0)
    _use_dsamp_prop = GetAttrProp("use_dsamp")
    use_dsamp = BoolItem(_("Use downsampling"), default=False).set_prop(
        "display", store=_use_dsamp_prop
    )
    dsamp_factor = IntItem(_("Downsampling factor"), default=10, min=1).set_prop(
        "display", active=_use_dsamp_prop
    )

    def update_param(self, curve: CurveItem) -> None:
        """Updates the parameters using values from a given CurveItem

        Args:
            curve: reference CurveItem instance
        """
        self.label = str(curve.title().text())
        self.symbol.update_param(curve.symbol())
        self.line.update_param(curve.pen())
        self.curvestyle = CURVESTYLE_NAME[curve.style()]
        self.baseline = curve.baseline()

    def update_item(self, curve: CurveItem) -> None:
        """Updates a given CurveItem using the current parameters

        Args:
            curve: instance of CurveItem to update
        """
        plot = curve.plot()
        if plot is not None:
            plot.blockSignals(True)  # Avoid unwanted calls of update_param
            # triggered by the setter methods below
        if not self._multiselection:
            # Non common parameters
            curve.setTitle(self.label)
        curve.setPen(self.line.build_pen())
        # Brush
        linecolor = QG.QColor(self.line.color)
        linecolor.setAlphaF(self.shade)
        brush = QG.QBrush(linecolor)
        if not self.shade:
            brush.setStyle(QC.Qt.NoBrush)
        curve.setBrush(brush)
        # Symbol
        self.symbol.update_symbol(curve)
        # Curve style, type and baseline
        curve.setStyle(getattr(QwtPlotCurve, self.curvestyle))
        curve.setBaseline(self.baseline)
        # Downsampling
        curve.update_data()
        if plot is not None:
            plot.blockSignals(False)


class CurveParam_MS(CurveParam):
    """Same as CurveParam but for multiselection"""

    _multiselection = True


ItemParameters.register_multiselection(CurveParam, CurveParam_MS)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Testing curve plot with data update
====================================

The purpose of this test is to implement a simple use case where a curve is
updated in a loop. The curve is created with random data and the values are
updated in a loop. The test is successful if the curve is updated in the
plot window.
"""

# guitest: show

from __future__ import annotations

import numpy as np
import pytest
from guidata.qthelpers import exec_dialog, qt_app_context
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from plotpy.builder import make
from plotpy.plot import PlotDialog, PlotOptions
from plotpy.tests import data as ptd


def get_data(variable_size: bool) -> tuple[np.ndarray, np.ndarray]:
    """Compute 1D Gaussian data and add a narrower Gaussian on top with a random
    position and amplitude."""
    size = np.random.randint(10, 200) if variable_size else 100
    amp = 0.3
    x, y = ptd.gen_1d_gaussian(size, sigma=10.0, x0=0.0, amp=amp)
    # Choose a random position: x0 has to be in the range [-10.0, 10.0]
    x0 = np.random.uniform(-10.0, 10.0)
    # Choose a random amplitude: a has to be in the range [0.1, 0.5]
    a = np.random.uniform(0.1, 0.7)
    # Add the narrower Gaussian on top
    y += ptd.gen_1d_gaussian(size, sigma=4.0, x0=x0, amp=a)[1]
    return x, y


class CurveUpdateDialog(PlotDialog):
    """Dialog box for curve update"""

    def __init__(
        self, title: str, variable_size: bool = False, auto_scale: bool = True
    ) -> None:
        self.variable_size = variable_size
        self.auto_scale = auto_scale
        options = PlotOptions(title="-", show_contrast=True, type="curve")
        super().__init__(title=title, toolbar=True, edit=False, options=options)
        self.resize(600, 600)
        self.timer = QC.QTimer()
        self.item = make.curve(
            *get_data(variable_size),
            color="blue",
            marker="o",
            markersize=5,
            markerfacecolor="cyan",
            markeredgecolor="blue",
        )
        plot = self.get_plot()
        plot.add_item(self.item)
        plot.set_active_item(self.item, select=False)
        self.counter = 0

    def populate_plot_layout(self) -> None:
        """Populate the plot layout with the item"""
        start_btn = QW.QPushButton("Start curve update")
        start_btn.clicked.connect(self.start_curve_update)
        self.add_widget(start_btn, 0, 0)
        stop_btn = QW.QPushButton("Stop curve update")
        stop_btn.clicked.connect(self.stop_curve_update)
        self.add_widget(stop_btn, 0, 1)
        variable_size_cb = QW.QCheckBox("Variable size")
        variable_size_cb.setChecked(self.variable_size)
        variable_size_cb.stateChanged.connect(self.toggle_variable_size)
        self.add_widget(variable_size_cb, 0, 2)
        auto_scale_cb = QW.QCheckBox("Auto scale")
        auto_scale_cb.setChecked(True)
        auto_scale_cb.stateChanged.connect(self.toggle_auto_scale)
        self.add_widget(auto_scale_cb, 0, 3)
        self.add_widget(self.plot_widget, 1, 0, 1, 0)

    def toggle_variable_size(self, state: int) -> None:
        """Toggle variable size"""
        self.variable_size = state == QC.Qt.Checked

    def toggle_auto_scale(self, state: int) -> None:
        """Toggle auto scale"""
        self.auto_scale = state == QC.Qt.Checked
        plot = self.get_plot()
        if self.auto_scale:
            plot.do_autoscale()

    def start_curve_update(self) -> None:
        """Start curve update"""
        self.timer.timeout.connect(self.update_curve)
        self.timer.start(100)

    def stop_curve_update(self) -> None:
        """Stop curve update"""
        self.timer.stop()

    def update_curve(self) -> None:
        """Update curve data"""
        data = get_data(self.variable_size)
        self.counter += 1
        plot = self.get_plot()
        self.item.set_data(data[0], data[1])
        plot.set_title(f"Curve update {self.counter:03d}")
        if self.auto_scale:
            plot.do_autoscale()
        else:
            plot.replot()

    def closeEvent(self, event: QG.QCloseEvent) -> None:
        """Close the dialog and stop the timer"""
        self.timer.stop()
        super().closeEvent(event)


@pytest.mark.skip(reason="Not relevant in automated test suite")
def test_curve_update() -> None:
    """Test curve update"""
    title = test_curve_update.__doc__
    with qt_app_context(exec_loop=False):
        dlg = CurveUpdateDialog(title)
        exec_dialog(dlg)


if __name__ == "__main__":
    test_curve_update()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""Run a command with environment variables loaded from a .env file."""

from __future__ import annotations

import os
import subprocess
import sys
from pathlib import Path


def load_env_file(env_path: str | None = None) -> None:
    """Load environment variables from a .env file."""
    # Set a flag to indicate that the environment has been loaded by this script
    # This prevents batch scripts (like utils.bat) from reloading .env and overwriting variables

    if env_path is None:
        # Get ".env" file from the current directory
        env_path = Path.cwd() / ".env"
    if not Path(env_path).is_file():
        raise FileNotFoundError(f"Environment file not found: {env_path}")
    print(f"Loading environment variables from: {env_path}")
    with open(env_path, encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#") or "=" not in line:
                continue
            key, value = line.split("=", 1)
            value = os.path.expandvars(value.strip())

            # Handle PATH variable specifically:
            # 1. Convert relative paths to absolute paths
            # 2. Normalize path separators
            if key.strip().upper() == "PATH":
                paths = value.split(os.pathsep)
                abs_paths = []
                for p in paths:
                    p = p.strip()
                    if not p:
                        continue
                    # Check if it looks like a relative path component
                    # (not starting with drive or root)
                    # Note: This simple check assumes standard usage in .env
                    if not os.path.isabs(p) and not p.startswith("%"):
                        try:
                            # Resolve relative to .env file directory
                            p = str((Path(env_path).parent / p).resolve())
                        except Exception:
                            pass  # Keep as is if resolution fails
                    abs_paths.append(os.path.normpath(p))
                value = os.pathsep.join(abs_paths)

            os.environ[key.strip()] = value
            print(f"  Loaded variable: {key.strip()}={value}")


def execute_command(command: list[str]) -> int:
    """Execute a command with the loaded environment variables."""
    print("Executing command:")
    print(" ".join(command))
    print("")
    result = subprocess.call(command)
    print(f"Process exited with code {result}")
    return result


def main() -> None:
    """Main function to load environment variables and execute a command."""
    if len(sys.argv) < 2:
        print("Usage: python run_with_env.py <command> [args ...]")
        sys.exit(1)
    print("🏃 Running with environment variables")
    load_env_file()
    return execute_command(sys.argv[1:])


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Resize dialog
-------------

The `resizedialog` module provides a dialog box providing essential GUI
for entering parameters needed to resize an image:
:py:class:`.widgets.resizedialog.ResizeDialog`.

Reference
~~~~~~~~~

.. autoclass:: ResizeDialog
   :members:
"""

from guidata.qthelpers import win32_fix_title_bar_background
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW
from qtpy.QtCore import Qt

from plotpy.config import _


def is_edit_valid(edit):
    """

    :param edit:
    :return:
    """
    text = edit.text()
    state = edit.validator().validate(text, 0)
    if isinstance(state, (tuple, list)):
        state = state[0]
    return state == QG.QIntValidator.Acceptable


class ResizeDialog(QW.QDialog):
    """Resize dialog box"""

    def __init__(self, parent, new_size, old_size, text="", keep_original_size=False):
        QW.QDialog.__init__(self, parent)
        win32_fix_title_bar_background(self)

        def intfunc(tup):
            return [int(val) for val in tup]

        if intfunc(new_size) == intfunc(old_size):
            self.keep_original_size = True
        else:
            self.keep_original_size = keep_original_size
        self.width, self.height = new_size
        self.old_width, self.old_height = old_size
        self.ratio = self.width / self.height

        layout = QW.QVBoxLayout()
        self.setLayout(layout)

        formlayout = QW.QFormLayout()
        layout.addLayout(formlayout)

        if text:
            label = QW.QLabel(text)
            label.setAlignment(Qt.AlignHCenter)
            formlayout.addRow(label)

        self.w_edit = w_edit = QW.QLineEdit(self)
        w_valid = QG.QIntValidator(w_edit)
        w_valid.setBottom(1)
        w_edit.setValidator(w_valid)

        self.h_edit = h_edit = QW.QLineEdit(self)
        h_valid = QG.QIntValidator(h_edit)
        h_valid.setBottom(1)
        h_edit.setValidator(h_valid)

        zbox = QW.QCheckBox(_("Original size"), self)

        formlayout.addRow(_("Width (pixels)"), w_edit)
        formlayout.addRow(_("Height (pixels)"), h_edit)
        formlayout.addRow("", zbox)

        formlayout.addRow(_("Original size:"), QW.QLabel("%d x %d" % old_size))
        self.z_label = QW.QLabel()
        formlayout.addRow(_("Zoom factor:"), self.z_label)

        # Button box
        self.bbox = bbox = QW.QDialogButtonBox(
            QW.QDialogButtonBox.Ok | QW.QDialogButtonBox.Cancel
        )
        bbox.accepted.connect(self.accept)
        bbox.rejected.connect(self.reject)
        layout.addWidget(bbox)

        self.w_edit.setText(str(int(self.width)))
        self.h_edit.setText(str(int(self.height)))
        self.update_widgets()

        self.setWindowTitle(_("Resize"))

        w_edit.textChanged.connect(self.width_changed)
        h_edit.textChanged.connect(self.height_changed)
        zbox.toggled.connect(self.toggled_no_zoom)
        zbox.setChecked(self.keep_original_size)

    def update_widgets(self):
        """ """
        valid = True
        for edit in (self.w_edit, self.h_edit):
            if not is_edit_valid(edit):
                valid = False
        self.bbox.button(QW.QDialogButtonBox.Ok).setEnabled(valid)
        self.z_label.setText("%d %s" % (100 * self.width / self.old_width, "%"))

    def width_changed(self, text):
        """

        :param text:
        """
        if is_edit_valid(self.sender()):
            self.width = int(text)
            self.height = int(self.width / self.ratio)
            self.h_edit.blockSignals(True)
            self.h_edit.setText(str(self.height))
            self.h_edit.blockSignals(False)
        self.update_widgets()

    def height_changed(self, text):
        """

        :param text:
        """
        if is_edit_valid(self.sender()):
            self.height = int(text)
            self.width = int(self.ratio * self.height)
            self.w_edit.blockSignals(True)
            self.w_edit.setText(str(self.width))
            self.w_edit.blockSignals(False)
        self.update_widgets()

    def toggled_no_zoom(self, state):
        """

        :param state:
        """
        self.keep_original_size = state
        if state:
            self.z_label.setText("100 %")
            self.bbox.button(QW.QDialogButtonBox.Ok).setEnabled(True)
        else:
            self.update_widgets()
        for widget in (self.w_edit, self.h_edit):
            widget.setDisabled(state)

    def get_zoom(self):
        """

        :return:
        """
        if self.keep_original_size:
            return 1
        elif self.width > self.height:
            return self.width / self.old_width
        else:
            return self.height / self.old_height
